# coding=utf-8
# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import time
import re

plugin_ctx = None


def send_yes(fsm_ctx):
    fsm_ctx.ctrl.sendline('yes')
    return True


def fpd_needs_upgd(ctx, fpd_location, fpd_type):

    """
    :param ctx
    :return: True or False

    Platform: NCS5500
    RP/0/RP0/CPU0:freta1.55#show hw-module fpd
                                                                       FPD Versions
                                                                   =================
    Location   Card type        HWver FPD device       ATR Status   Running Programd
    ------------------------------------------------------------------------------
    0/0       NC55-18H18F       1.0   MIFPGA               CURRENT    0.03    0.03
    0/0       NC55-18H18F       1.0   Bootloader           CURRENT    1.11    1.11
    0/0       NC55-18H18F       1.0   IOFPGA               CURRENT    0.20    0.20
    0/1       NC55-24X100G-SE   1.0   MIFPGA               CURRENT    0.03    0.03
    0/1       NC55-24X100G-SE   1.0   Bootloader           NEED UPGD  1.09    1.09
    0/1       NC55-24X100G-SE   1.0   IOFPGA               NEED UPGD  0.08    0.08
    0/7       NC55-36X100G-S    1.0   MIFPGA               CURRENT    0.06    0.06
    0/7       NC55-36X100G-S    1.0   Bootloader           CURRENT    1.11    1.11
    0/7       NC55-36X100G-S    1.0   IOFPGA               CURRENT    0.09    0.09
    0/RP0     NC55-RP           1.1   Bootloader           NEED UPGD  9.21    9.21
    0/RP0     NC55-RP           1.1   IOFPGA               NEED UPGD  0.08    0.08
    0/RP1     NC55-RP           1.1   Bootloader           NEED UPGD  9.21    9.21
    0/RP1     NC55-RP           1.1   IOFPGA               NEED UPGD  0.08    0.08
    0/FC0     NC55-5508-FC      1.0   Bootloader           CURRENT    1.70    1.70
    0/FC0     NC55-5508-FC      1.0   IOFPGA               NEED UPGD  0.15    0.15
    0/FC1     NC55-5508-FC      1.0   Bootloader           NEED UPGD  1.65    1.65
    0/FC1     NC55-5508-FC      1.0   IOFPGA               NEED UPGD  0.13    0.13
    0/FC3     NC55-5508-FC      1.0   Bootloader           NEED UPGD  1.65    1.65
    0/FC3     NC55-5508-FC      1.0   IOFPGA               NEED UPGD  0.13    0.13
    0/FC5     NC55-5508-FC      1.0   Bootloader           NEED UPGD  1.65    1.65
    0/FC5     NC55-5508-FC      1.0   IOFPGA               NEED UPGD  0.13    0.13
    0/SC0     NC55-SC           1.4   Bootloader           NEED UPGD  1.65    1.65
    0/SC0     NC55-SC           1.4   IOFPGA               NEED UPGD  0.07    0.07
    0/SC1     NC55-SC           1.4   Bootloader           NEED UPGD  1.65    1.65
    0/SC1     NC55-SC           1.4   IOFPGA               NEED UPGD  0.07    0.07
    """

    need_upgd = []
    sl = ['Card', 'FPD', 'ATR', 'Status', 'Running']
    dl = {}

    output = ctx.send("show hw-module fpd", timeout=600)
    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:
        line = line.strip()

        if line[0:8] == 'Location':
            for s in sl:
                n = line.find(s)
                if n != -1:
                    if s == 'Card':
                        n -= 1
                    dl[str(s)] = n
                else:
                    ctx.error("unrecognized show hw-module fpd header {}".format(line))
                    return False

        if line[0].isdigit():
            status = line[dl['Status']:dl['Running']].strip()
            location = line[0:dl['Card']].strip()
            type = line[dl['FPD']:dl['ATR']].strip()
            if fpd_location == 'all' and fpd_type == 'all':
                if 'NEED UPGD'in status:
                    need_upgd.append(location)
            elif fpd_location and not fpd_type:
                if location == fpd_location:
                    if 'NEED UPGD'in status:
                        need_upgd.append(location)
                else:
                    continue
            elif not fpd_location and fpd_type:
                if type == fpd_type:
                    if 'NEED UPGD' in status:
                        need_upgd.append(location)
                else:
                    continue
            else:
                if location == fpd_location and type == fpd_type:
                    if 'NEED UPGD' in status:
                        need_upgd.append(location)
                        break
                else:
                    continue

    need_upgd = list(set(need_upgd))
    return need_upgd


def fpd_is_current(ctx, fpd_location, fpd_type):
    """
    :param ctx
    :return: True or False

    Platform: NCS5500
    RP/0/RP0/CPU0:freta1.55#show hw-module fpd
                                                                   FPD Versions
                                                                   =================
    Location   Card type        HWver FPD device       ATR Status   Running Programd
    ------------------------------------------------------------------------------
    0/0       NC55-18H18F       1.0   MIFPGA               CURRENT    0.03    0.03
    0/0       NC55-18H18F       1.0   Bootloader           CURRENT    1.11    1.11
    0/0       NC55-18H18F       1.0   IOFPGA               CURRENT    0.20    0.20
    0/1       NC55-24X100G-SE   1.0   MIFPGA               CURRENT    0.03    0.03
    0/1       NC55-24X100G-SE   1.0   Bootloader           CURRENT    1.11    1.11
    0/1       NC55-24X100G-SE   1.0   IOFPGA               CURRENT    0.12    0.12
    0/7       NC55-36X100G-S    1.0   MIFPGA               CURRENT    0.06    0.06
    0/7       NC55-36X100G-S    1.0   Bootloader           CURRENT    1.11    1.11
    0/7       NC55-36X100G-S    1.0   IOFPGA               CURRENT    0.09    0.09
    0/RP0     NC55-RP           1.1   Bootloader           RLOAD REQ  9.21    9.23
    0/RP0     NC55-RP           1.1   IOFPGA               RLOAD REQ  0.08    0.09
    0/RP1     NC55-RP           1.1   Bootloader           RLOAD REQ  9.21    9.23
    0/RP1     NC55-RP           1.1   IOFPGA               RLOAD REQ  0.08    0.09
    0/FC0     NC55-5508-FC      1.0   Bootloader           CURRENT    1.70    1.70
    0/FC0     NC55-5508-FC      1.0   IOFPGA               RLOAD REQ  0.15    0.16
    0/FC1     NC55-5508-FC      1.0   Bootloader           RLOAD REQ  1.65    1.70
    0/FC1     NC55-5508-FC      1.0   IOFPGA               RLOAD REQ  0.13    0.16
    0/FC3     NC55-5508-FC      1.0   Bootloader           RLOAD REQ  1.65    1.70
    0/FC3     NC55-5508-FC      1.0   IOFPGA               RLOAD REQ  0.13    0.16
    0/FC5     NC55-5508-FC      1.0   Bootloader           RLOAD REQ  1.65    1.70
    0/FC5     NC55-5508-FC      1.0   IOFPGA               RLOAD REQ  0.13    0.16
    0/SC0     NC55-SC           1.4   Bootloader           RLOAD REQ  1.65    1.70
    0/SC0     NC55-SC           1.4   IOFPGA               RLOAD REQ  0.07    0.08
    0/SC1     NC55-SC           1.4   Bootloader           RLOAD REQ  1.65    1.70
    0/SC1     NC55-SC           1.4   IOFPGA               RLOAD REQ  0.07    0.08
    """

    is_current = True
    sl = ['Card', 'FPD', 'ATR', 'Status', 'Running']
    dl = {}

    output = ctx.send("show hw-module fpd", timeout=600)
    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:
        line = line.strip()

        if line[0:8] == 'Location':
            for s in sl:
                n = line.find(s)
                if n != -1:
                    if s == 'Card':
                        n -= 1
                    dl[str(s)] = n
                else:
                    ctx.error("unrecognized show hw-module fpd header {}".format(line))
                    return False

        if line[0].isdigit():
            status = line[dl['Status']:dl['Running']].strip()
            location = line[0:dl['Card']].strip()
            type = line[dl['FPD']:dl['ATR']].strip()
            if 'N/A' in status:
                continue
            if fpd_location == 'all' and fpd_type == 'all':
                if 'CURRENT' not in status:
                    is_current = False
                    break
            elif fpd_location and not fpd_type:
                if location == fpd_location:
                    if 'CURRENT' not in status:
                        is_current = False
                        break
                else:
                    continue
            elif not fpd_location and fpd_type:
                if type == fpd_type:
                    if 'CURRENT' not in status:
                        is_current = False
                        break
                else:
                    continue
            else:
                if location == fpd_location and type == fpd_type:
                    if 'CURRENT' not in status:
                        is_current = False
                        break
                else:
                    continue

    return is_current


def fpd_needs_reload(ctx, fpd_location, fpd_type):
    """
    :param ctx
    :return: True or False

    Platform: NCS5500
    RP/0/RP0/CPU0:freta1.55#show hw-module fpd
                                                                   FPD Versions
                                                                   =================
    Location   Card type        HWver FPD device       ATR Status   Running Programd
    ------------------------------------------------------------------------------
    0/0       NC55-18H18F       1.0   MIFPGA               CURRENT    0.03    0.03
    0/0       NC55-18H18F       1.0   Bootloader           CURRENT    1.11    1.11
    0/0       NC55-18H18F       1.0   IOFPGA               CURRENT    0.20    0.20
    0/1       NC55-24X100G-SE   1.0   MIFPGA               CURRENT    0.03    0.03
    0/1       NC55-24X100G-SE   1.0   Bootloader           CURRENT    1.11    1.11
    0/1       NC55-24X100G-SE   1.0   IOFPGA               CURRENT    0.12    0.12
    0/7       NC55-36X100G-S    1.0   MIFPGA               CURRENT    0.06    0.06
    0/7       NC55-36X100G-S    1.0   Bootloader           CURRENT    1.11    1.11
    0/7       NC55-36X100G-S    1.0   IOFPGA               CURRENT    0.09    0.09
    0/RP0     NC55-RP           1.1   Bootloader           RLOAD REQ  9.21    9.23
    0/RP0     NC55-RP           1.1   IOFPGA               RLOAD REQ  0.08    0.09
    0/RP1     NC55-RP           1.1   Bootloader           RLOAD REQ  9.21    9.23
    0/RP1     NC55-RP           1.1   IOFPGA               RLOAD REQ  0.08    0.09
    0/FC0     NC55-5508-FC      1.0   Bootloader           CURRENT    1.70    1.70
    0/FC0     NC55-5508-FC      1.0   IOFPGA               RLOAD REQ  0.15    0.16
    0/FC1     NC55-5508-FC      1.0   Bootloader           RLOAD REQ  1.65    1.70
    0/FC1     NC55-5508-FC      1.0   IOFPGA               RLOAD REQ  0.13    0.16
    0/FC3     NC55-5508-FC      1.0   Bootloader           RLOAD REQ  1.65    1.70
    0/FC3     NC55-5508-FC      1.0   IOFPGA               RLOAD REQ  0.13    0.16
    0/FC5     NC55-5508-FC      1.0   Bootloader           RLOAD REQ  1.65    1.70
    0/FC5     NC55-5508-FC      1.0   IOFPGA               RLOAD REQ  0.13    0.16
    0/SC0     NC55-SC           1.4   Bootloader           RLOAD REQ  1.65    1.70
    0/SC0     NC55-SC           1.4   IOFPGA               RLOAD REQ  0.07    0.08
    0/SC1     NC55-SC           1.4   Bootloader           RLOAD REQ  1.65    1.70
    0/SC1     NC55-SC           1.4   IOFPGA               RLOAD REQ  0.07    0.08

    Platform: NCS1K
    RP/0/RP0/CPU0:PROD_19#show hw fpd
    Fri Jul 14 02:56:06.415 IST
                                                                   FPD Versions
                                                                   =================
    Location   Card type        HWver FPD device       ATR Status   Running Programd
    ------------------------------------------------------------------------------
    0/0       NCS1002-K9        1.2   CDSP_PORT_05         CURRENT    3.76    3.76
    0/0       NCS1002-K9        1.2   CDSP_PORT_06         CURRENT    3.76    3.76
    0/0       NCS1002-K9        1.2   CDSP_PORT_12         CURRENT    3.76    3.76
    0/0       NCS1002-K9        1.2   CDSP_PORT_13         CURRENT    3.76    3.76
    0/0       NCS1002-K9        1.2   CDSP_PORT_19         CURRENT    3.76    3.76
    0/0       NCS1002-K9        1.2   CDSP_PORT_20         CURRENT    3.76    3.76
    0/0       NCS1002-K9        1.2   CDSP_PORT_26         CURRENT    3.76    3.76
    0/0       NCS1002-K9        1.2   CDSP_PORT_27         CURRENT    3.76    3.76
    0/0       NCS1002-K9              CFP2_PORT_05         NOT READY
    0/0       NCS1002-K9        2.1   CFP2_PORT_06         CURRENT    5.23    5.23
    0/0       NCS1002-K9              CFP2_PORT_12         NOT READY
    0/0       NCS1002-K9        2.1   CFP2_PORT_13         CURRENT    5.23    5.23
    0/0       NCS1002-K9              CFP2_PORT_19         NOT READY
    0/0       NCS1002-K9              CFP2_PORT_20         NOT READY
    0/0       NCS1002-K9              CFP2_PORT_26         NOT READY
    0/0       NCS1002-K9        2.1   CFP2_PORT_27         CURRENT    5.23    5.23
    0/0       NCS1002-K9        0.1   CTRL_BKP_LOW     B   CURRENT            2.23
    0/0       NCS1002-K9        0.1   CTRL_BKP_UP      B   CURRENT            2.23
    0/0       NCS1002-K9        0.1   CTRL_FPGA_LOW        CURRENT    2.23    2.23
    0/0       NCS1002-K9        0.1   CTRL_FPGA_UP         CURRENT    2.23    2.23
    0/RP0     NCS1K-CNTLR       0.1   BIOS_Backup      BS  RLOAD REQ         13.70
    0/RP0     NCS1K-CNTLR       0.1   BIOS_Primary      S  RLOAD REQ 13.40   13.70
    0/RP0     NCS1K-CNTLR       0.1   Daisy_Duke_BKP   BS  CURRENT            0.15
    0/RP0     NCS1K-CNTLR       0.1   Daisy_Duke_FPGA   S  RLOAD REQ  0.15    0.17
    0/PM0     NCS1K-2KW-AC      0.0   PO-PriMCU            CURRENT
    0/PM1     NCS1K-2KW-AC      0.0   PO-PriMCU            CURRENT    4.00    4.00

    Platform: ASR9K-X64
    RP/0/RSP0/CPU0:vkg3#show hw-module fpd
                                                                       FPD Versions
                                                                    ===============
    Location  Card type         HWver FPD device       ATR Status   Run    Programd
    -------------------------------------------------------------------------------
    0/0       A99-8X100GE-SE    1.0   CBC                  CURRENT   38.23   38.23
    0/0       A99-8X100GE-SE    1.0   Dalla                CURRENT    1.09    1.09
    0/0       A99-8X100GE-SE    1.0   IPU-FPGA             CURRENT    1.91    1.91
    0/0       A99-8X100GE-SE    1.0   IPU-FSBL             CURRENT    1.96    1.96
    0/0       A99-8X100GE-SE    1.0   IPU-Linux            CURRENT    1.96    1.96
    0/0       A99-8X100GE-SE    1.0   Meldun-0             CURRENT    1.07    1.07
    0/0       A99-8X100GE-SE    1.0   Meldun-1             CURRENT    1.07    1.07
    0/0       A99-8X100GE-SE    1.0   Primary-BIOS         CURRENT    8.43    8.43
    0/RP0     A99-RP2-SE        1.0   Alpha-FPGA           CURRENT    0.16    0.16
    0/RP0     A99-RP2-SE        1.0   CBC-0                CURRENT   35.12   35.12
    0/RP0     A99-RP2-SE        1.0   CBC-1                CURRENT   35.12   35.12
    0/RP0     A99-RP2-SE        1.0   Cha-FPGA             CURRENT    0.05    0.05
    0/RP0     A99-RP2-SE        1.0   IPU-FPGA             CURRENT    0.52    0.52
    0/RP0     A99-RP2-SE        1.0   IPU-FSBL             CURRENT    1.96    1.96
    0/RP0     A99-RP2-SE        1.0   IPU-Linux            CURRENT    1.96    1.96
    0/RP0     A99-RP2-SE        1.0   Omega-FPGA           CURRENT    0.15    0.15
    0/RP0     A99-RP2-SE        1.0   Optimus-FPGA         CURRENT    0.12    0.12
    0/RP0     A99-RP2-SE        1.0   Primary-BIOS         CURRENT   14.28   14.28
    0/RP1     A99-RP2-SE        1.0   Alpha-FPGA           CURRENT    0.16    0.16
    0/RP1     A99-RP2-SE        1.0   CBC-0                CURRENT   35.12   35.12
    0/RP1     A99-RP2-SE        1.0   CBC-1                CURRENT   35.12   35.12
    0/RP1     A99-RP2-SE        1.0   Cha-FPGA             CURRENT    0.05    0.05
    0/RP1     A99-RP2-SE        1.0   IPU-FPGA             CURRENT    0.52    0.52
    0/RP1     A99-RP2-SE        1.0   IPU-FSBL             CURRENT    1.96    1.96
    0/RP1     A99-RP2-SE        1.0   IPU-Linux            CURRENT    1.96    1.96
    0/RP1     A99-RP2-SE        1.0   Omega-FPGA           CURRENT    0.15    0.15
    0/RP1     A99-RP2-SE        1.0   Optimus-FPGA         CURRENT    0.12    0.12
    0/RP1     A99-RP2-SE        1.0   Primary-BIOS         CURRENT   14.28   14.28
    0/FC0     A99-SFC2          1.0   CBC                  CURRENT   37.20   37.20
    0/FC0     A99-SFC2          1.0   IPU-FPGA             CURRENT    0.33    0.33
    0/FC0     A99-SFC2          1.0   IPU-FSBL             CURRENT    1.79    1.79
    0/FC0     A99-SFC2          1.0   IPU-Linux            CURRENT    1.79    1.79
    0/FC1     A99-SFC2          1.0   CBC                  CURRENT   37.20   37.20
    0/FC1     A99-SFC2          1.0   IPU-FPGA             CURRENT    0.33    0.33
    0/FC1     A99-SFC2          1.0   IPU-FSBL             CURRENT    1.79    1.79
    0/FC1     A99-SFC2          1.0   IPU-Linux            CURRENT    1.79    1.79
    0/FC2     A99-SFC2          1.0   CBC                  CURRENT   37.20   37.20
    0/FC2     A99-SFC2          1.0   IPU-FPGA             CURRENT    0.33    0.33
    0/FC2     A99-SFC2          1.0   IPU-FSBL             CURRENT    1.79    1.79
    0/FC2     A99-SFC2          1.0   IPU-Linux            CURRENT    1.79    1.79
    0/FC3     A99-SFC2          1.0   CBC                  CURRENT   37.20   37.20
    0/FC3     A99-SFC2          1.0   IPU-FPGA             CURRENT    0.33    0.33
    0/FC3     A99-SFC2          1.0   IPU-FSBL             CURRENT    1.79    1.79
    0/FC3     A99-SFC2          1.0   IPU-Linux            CURRENT    1.79    1.79
    0/FC4     A99-SFC2          1.0   CBC                  CURRENT   37.20   37.20
    0/FC4     A99-SFC2          1.0   IPU-FPGA             CURRENT    0.33    0.33
    0/FC4     A99-SFC2          1.0   IPU-FSBL             CURRENT    1.79    1.79
    0/FC4     A99-SFC2          1.0   IPU-Linux            CURRENT    1.79    1.79
    0/FT0     ASR-9912-FAN      1.0   CBC                  CURRENT   31.05   31.05
    0/FT1     ASR-9912-FAN      1.0   CBC                  CURRENT   31.05   31.05
    0/PT1     PWR-3KW-AC-V2     2.0   PM0-DT-PriMCU        UPGD SKIP  6.01    6.01
    0/PT1     PWR-3KW-AC-V2     2.0   PM0-DT-Sec54vMCU     UPGD SKIP  6.01    6.01
    0/PT1     PWR-3KW-AC-V2     2.0   PM0-DT-Sec5vMCU      UPGD SKIP  6.03    6.03
    0/PT2     PWR-3KW-AC-V2     2.0   PM0-DT-PriMCU        UPGD SKIP  6.01    6.01
    0/PT2     PWR-3KW-AC-V2     2.0   PM0-DT-Sec54vMCU     UPGD SKIP  6.01    6.01
    0/PT2     PWR-3KW-AC-V2     2.0   PM0-DT-Sec5vMCU      UPGD SKIP  6.03    6.03
    0/BPID0   ASR-9912-AC       1.0   CBC                  CURRENT    7.105   7.105
    """

    reload_ready = True
    sl = ['Card', 'FPD', 'ATR', 'Status', 'Running']
    dl = {}

    output = ctx.send("show hw-module fpd", timeout=600)
    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:
        line = line.strip()

        if line[0:8] == 'Location':
            for s in sl:
                n = line.find(s)
                if n != -1:
                    if s == 'Card':
                        n -= 1
                    dl[str(s)] = n
                else:
                    ctx.error("unrecognized show hw-module fpd header {}".format(line))
                    return False

        if line[0].isdigit():
            status = line[dl['Status']:dl['Running']].strip()
            location = line[0:dl['Card']].strip()
            type = line[dl['FPD']:dl['ATR']].strip()
            if 'N/A' in status:
                continue
            # take care of NSR1K CFP2 exception
            if 'NOT READY' in status:
                version = line[dl['Running']:].strip()
                if not version:
                    continue
            # take care of asr9k-x64 exception
            if 'UPGD SKIP' in status:
                continue
            if fpd_location == 'all' and fpd_type == 'all':
                if 'CURRENT' not in status and 'RLOAD REQ' not in status:
                    reload_ready = False
                    break
            elif fpd_location and fpd_type == 'all':
                if location == fpd_location:
                    if 'CURRENT' not in status and 'RLOAD REQ' not in status:
                        reload_ready = False
                        break
                else:
                    continue
            else:
                if location == fpd_location and type == fpd_type:
                    if 'CURRENT' not in status and 'RLOAD REQ' not in status:
                        reload_ready = False
                        break
                else:
                    continue

    return reload_ready


def fpd_check_status(ctx, fpd_location, fpd_type):
    """
    :param ctx
    :return: True or False

    Platform: NCS5500
    RP/0/RP0/CPU0:xrg-ncs-07#show hw-module fpd
    Wed Jun 28 17:56:16.811 UTC
                                                                   FPD Versions
                                                                   =================
    Location   Card type        HWver FPD device       ATR Status   Running Programd
    ------------------------------------------------------------------------------
    0/RP0     NCS-5501          1.0   MB-MIFPGA            CURRENT    1.01    1.01
    0/RP0     NCS-5501          1.0   Bootloader           CURRENT    1.13    1.13
    0/RP0     NCS-5501          1.0   CPU-IOFPGA           CURRENT    1.14    1.14
    0/RP0     NCS-5501          1.0   MB-IOFPGA            CURRENT    1.04    1.04

    Platform: ASR9K-X64
    RP/0/RSP0/CPU0:vkg3#show hw-module fpd
    Wed Jun 28 11:05:35.914 UTC
                                                                   FPD Versions
                                                                   =================
    Location   Card type        HWver FPD device       ATR Status   Running Programd
    ------------------------------------------------------------------------------
    0/RSP0    A9K-RSP880-SE     1.0   Alpha-FPGA           CURRENT    0.16    0.16
    0/RSP0    A9K-RSP880-SE     1.0   CBC                  CURRENT   34.38   34.38
    0/RSP0    A9K-RSP880-SE     1.0   Cha-FPGA             CURRENT    0.07    0.07
    0/RSP0    A9K-RSP880-SE     1.0   IPU-FPGA             CURRENT    0.56    0.56
    0/RSP0    A9K-RSP880-SE     1.0   IPU-FSBL             CURRENT    1.101   1.101
    0/RSP0    A9K-RSP880-SE     1.0   IPU-Linux            CURRENT    1.101   1.101
    0/RSP0    A9K-RSP880-SE     1.0   Omega-FPGA           CURRENT    0.16    0.16
    0/RSP0    A9K-RSP880-SE     1.0   Optimus-FPGA         CURRENT    0.12    0.12
    0/RSP0    A9K-RSP880-SE     1.0   Primary-BIOS         CURRENT   10.58   10.58
    0/RSP1    A9K-RSP880-SE     1.0   Alpha-FPGA           CURRENT    0.16    0.16
    0/RSP1    A9K-RSP880-SE     1.0   CBC                  CURRENT   34.38   34.38
    0/RSP1    A9K-RSP880-SE     1.0   Cha-FPGA             CURRENT    0.07    0.07
    0/RSP1    A9K-RSP880-SE     1.0   IPU-FPGA             CURRENT    0.56    0.56
    0/RSP1    A9K-RSP880-SE     1.0   IPU-FSBL             CURRENT    1.101   1.101
    0/RSP1    A9K-RSP880-SE     1.0   IPU-Linux            CURRENT    1.101   1.101
    0/RSP1    A9K-RSP880-SE     1.0   Omega-FPGA           CURRENT    0.16    0.16
    0/RSP1    A9K-RSP880-SE     1.0   Optimus-FPGA         CURRENT    0.12    0.12
    0/RSP1    A9K-RSP880-SE     1.0   Primary-BIOS         CURRENT   10.58   10.58
    0/FT0     ASR-9904-FAN      1.0   CBC                  CURRENT   31.05   31.05
    0/0       A9K-8X100GE-SE    1.0   CBC                  CURRENT   38.23   38.23
    0/0       A9K-8X100GE-SE    1.0   Dalla                CURRENT    1.09    1.09
    0/0       A9K-8X100GE-SE    1.0   IPU-FPGA             CURRENT    1.93    1.93
    0/0       A9K-8X100GE-SE    1.0   IPU-FSBL             CURRENT    1.100   1.100
    0/0       A9K-8X100GE-SE    1.0   IPU-Linux            CURRENT    1.100   1.100
    0/0       A9K-8X100GE-SE    1.0   Meldun-0             CURRENT    1.07    1.07
    0/0       A9K-8X100GE-SE    1.0   Meldun-1             CURRENT    1.07    1.07
    0/0       A9K-8X100GE-SE    1.0   Primary-BIOS         CURRENT    8.43    8.43
    0/BPID0   ASR-9904-AC       1.0   CBC                  CURRENT    7.105   7.105
    0/PT0     PWR-3KW-AC-V2     3.0   PM0-EM-PriMCU        N/A        3.06    3.06
    0/PT0     PWR-3KW-AC-V2     3.0   PM0-EM-Sec54vMCU     CURRENT    3.12    3.12
    0/PT0     PWR-3KW-AC-V2     3.0   PM0-EM-Sec5vMCU      CURRENT    3.18    3.18
    0/PT0     PWR-3KW-AC-V2     3.0   PM1-EM-PriMCU        N/A        3.06    3.06
    0/PT0     PWR-3KW-AC-V2     3.0   PM1-EM-Sec54vMCU     CURRENT    3.12    3.12
    0/PT0     PWR-3KW-AC-V2     3.0   PM1-EM-Sec5vMCU      CURRENT    3.18    3.18
    0/PT0     PWR-3KW-AC-V2     3.0   PM2-EM-PriMCU        N/A        3.06    3.06
    0/PT0     PWR-3KW-AC-V2     3.0   PM2-EM-Sec54vMCU     CURRENT    3.12    3.12
    0/PT0     PWR-3KW-AC-V2     3.0   PM2-EM-Sec5vMCU      CURRENT    3.18    3.18
    0/PT0     PWR-3KW-AC-V2     3.0   PM3-EM-PriMCU        N/A        3.06    3.06
    0/PT0     PWR-3KW-AC-V2     3.0   PM3-EM-Sec54vMCU     CURRENT    3.12    3.12
    0/PT0     PWR-3KW-AC-V2     3.0   PM3-EM-Sec5vMCU      CURRENT    3.18    3.18
    """

    current_status = True
    sl = ['Card', 'FPD', 'ATR', 'Status', 'Running']
    dl = {}

    output = ctx.send("show hw-module fpd", timeout=600)
    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:
        line = line.strip()

        if line[0:8] == 'Location':
            for s in sl:
                n = line.find(s)
                if n != -1:
                    if s == 'Card':
                        n -= 1
                    dl[str(s)] = n
                else:
                    ctx.error("unrecognized show hw-module fpd header {}".format(line))
                    return False

        if line[0].isdigit():
            status = line[dl['Status']:dl['Running']].strip()
            location = line[0:dl['Card']].strip()
            type = line[dl['FPD']:dl['ATR']].strip()
            if 'N/A' in status:
                ctx.warning('FPD Status: {}'.format(line))
                continue
            # take care of NSR1K CFP2 exception
            if 'NOT READY' in status:
                version = line[dl['Running']:].strip()
                if not version:
                    continue
            # take care of asr9k-x64 exception
            if 'UPGD SKIP' in status:
                continue
            if fpd_location == 'all' and fpd_type == 'all':
                if 'CURRENT' not in status:
                    ctx.warning('FPD Status: {}'.format(line))
                    current_status = False
                    break
            elif fpd_location and not fpd_type:
                if location == fpd_location:
                    if 'CURRENT' not in status:
                        current_status = False
                        break
                else:
                    continue
            elif not fpd_location and fpd_type:
                if type == fpd_type:
                    if 'CURRENT' not in status:
                        current_status = False
                        break
                else:
                    continue
            else:
                if location == fpd_location and type == fpd_type:
                    if 'CURRENT' not in status:
                        current_status = False
                        break
                else:
                    continue

    return current_status


def hw_fpd_upgd(ctx, location, type):
    """
    :param ctx
    :return: True or False

    Platform: NCS5500
    RP/0/RP0/CPU0:freta1.55#upgrade hw-module location all fpd all
    Wed May 24 09:14:00.817 CEST
    upgrade command issued (use "show hw-module fpd" to check upgrade status)
    """

    cmd = 'upgrade hw-module location ' + location + ' fpd ' + type
    output = ctx.send(cmd)

    if 'upgrade command issued' in output:
        return True
    else:
        return False


def hw_fpd_reload(ctx, location):
    """
    :param ctx
    :return: True or False

    Platform: NCS5500
    RP/0/RP0/CPU0:xrg-ncs-07#admin hw-module location all reload
    Tue Jun 27 00:47:15.279 UTC

    lab connected from 127.0.0.1 using console on xr-vm_node0_RP0_CPU0
    sysadmin-vm:0_RP0# terminal length 0
    Tue Jun  27 00:47:16.337 UTC
    sysadmin-vm:0_RP0# hw-module location all reload
    Tue Jun  27 00:47:16.379 UTC
    Reload hardware module ? [no,yes] yes
    yes
    result Card graceful reload request on all succeeded.
    """
    global plugin_ctx
    plugin_ctx = ctx

    cmd = 'admin hw-module location ' + location + ' reload'

    # Seeing this message without the reboot prompt indicates a non-reload situation
    RELOAD_MODULE = re.compile(r"Reload hardware module \? \[no,yes\]")
    HOST_PROMPT = re.compile(ctx.prompt)

    events = [RELOAD_MODULE, HOST_PROMPT]
    transitions = [
        (RELOAD_MODULE, [0], 1, send_yes, 60),
        (HOST_PROMPT, [1], -1, None, 120)
    ]

    if not ctx.run_fsm("hw-module reload", cmd, events, transitions, timeout=120):
        ctx.warning("Failed: {}".format(cmd))
        return False

    return True


def wait_for_fpd_upgd(ctx, location, type):
    """
    :param ctx
    :return: True or False
    """
    begin = time.time()

    ctx.info("FPD upgrade in progress")

    timeout = 7200
    poll_time = 30
    time_waited = 0

    ctx.info("Waiting for FPD upgrade to complete")
    ctx.post_status("Waiting for FPD upgrade to complete")

    while 1:
        # Wait till all nodes are in CURRENT or RLOAD REQ
        time.sleep(poll_time)

        if fpd_needs_reload(ctx, location, type):
            ctx.info("Location = {}, FPD device = {} in desired state.".format(location, type))
            elapsed = time.time() - begin
            ctx.info("Overall fpd upgrade time: {} minute(s) {:.0f} second(s)".format(elapsed // 60, elapsed % 60))
            return True

        time_waited += poll_time
        if time_waited >= timeout:
            break

    # Some nodes did not come to CURRENT or RLOAD REQ
    ctx.error("FPD Upgrade Timeout: some nodes are not in CURRENT or RLOAD REQ")
    # this will never be executed
    return False



# =============================================================================
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================
import re


def parse_show_platform(ctx, output):
    """
    :param ctx: PluginContext
    :param output: output from 'show platform'
    :return: dictionary of nodes

    ASR9K:
    RP/0/RSP0/CPU0:PE4#admin show platform
    Node            Type                      State            Config State
    -----------------------------------------------------------------------------
    0/RSP0/CPU0     A9K-RSP880-LT-SE(Active)  IOS XR RUN       PWR,NSHUT,MON
    0/RSP1/CPU0     A9K-RSP880-LT-SE(Standby) IOS XR RUN       PWR,NSHUT,MON
    0/FT0/SP        ASR-9006-FAN-V2           READY
    0/FT1/SP        ASR-9006-FAN-V2           READY
    0/0/CPU0        A9K-24X10GE-1G-SE         IOS XR RUN       PWR,NSHUT,MON
    0/PS0/M0/SP     A9K-3KW-AC                READY            PWR,NSHUT,MON
    0/PS0/M1/SP     A9K-3KW-AC                READY            PWR,NSHUT,MON
    0/PS0/M2/SP     A9K-3KW-AC                READY            PWR,NSHUT,MON

    XR12K:
    RP/0/7/CPU0:GSR-PE19#admin show platform
    Node		    Type		    PLIM		    State		    Config State
    -----------------------------------------------------------------------------
    0/0/CPU0        L3LC Eng 5      Jacket Card     IOS XR RUN      PWR,NSHUT,MON
    0/1/CPU0        L3LC Eng 5+     Jacket Card     IOS XR RUN      PWR,NSHUT,MON
    0/1/0           SPA             SPA-10X1GE-V2   READY           PWR,NSHUT
    0/1/1           SPA             SPA-1XCHOC48/DS READY           PWR,NSHUT
    0/3/CPU0        L3LC Eng 5+     Jacket Card     IOS XR RUN      PWR,NSHUT,MON
    0/3/0           SPA             SPA-10X1GE-V2   READY           PWR,NSHUT
    0/5/CPU0        L3LC Eng 3      OC12-ATM-4      IOS XR RUN      PWR,NSHUT,MON
    0/6/CPU0        L3LC Eng 5      Jacket Card     IOS XR RUN      PWR,NSHUT,MON
    0/17/CPU0       CSC_OC192(P)    N/A             PWD             PWR,NSHUT,MON
    0/18/CPU0       SFC_OC192       N/A             PWD             PWR,NSHUT,MON
    0/19/CPU0       SFC_OC192       N/A             PWD             PWR,NSHUT,MON

    CRS:
    Node          Type              PLIM               State           Config State
    ------------- ----------------- ------------------ --------------- ---------------
    0/0/CPU0      MSC-X             40-10GbE           IOS XR RUN      PWR,NSHUT,MON
    0/1/SP        MSC-B(SP)         N/A                IOS XR RUN      PWR,NSHUT,MON
    0/1/CPU0      MSC-B             Jacket Card        IOS XR RUN      PWR,NSHUT,MON
    0/1/1         MSC-B(SPA)        1x10GE             OK              PWR,NSHUT,MON
    0/1/2         MSC-B(SPA)        10X1GE             OK              PWR,NSHUT,MON
    0/2/CPU0      FP-X              4-100GbE           IOS XR RUN      PWR,NSHUT,MON
    0/3/CPU0      MSC-140G          N/A                UNPOWERED       NPWR,NSHUT,MON
    0/4/CPU0      FP-X              N/A                UNPOWERED       NPWR,NSHUT,MON
    0/7/CPU0      MSC-X             40-10GbE           IOS XR RUN      PWR,NSHUT,MON
    0/8/CPU0      MSC-140G          N/A                UNPOWERED       NPWR,NSHUT,MON
    0/14/CPU0     MSC-X             4-100GbE           IOS XR RUN      PWR,NSHUT,MON
    0/RP0/CPU0    RP(Active)        N/A                IOS XR RUN      PWR,NSHUT,MON
    0/RP1/CPU0    RP(Standby)       N/A                IOS XR RUN      PWR,NSHUT,MON

    """
    host = ctx.get_host
    inventory = {}

    # Use the tuple to record the begin and end positions of each column.  Unfortunately,
    # we cannot rely on the header line as it contains a mixed of space and tab characters.
    if host.family == 'ASR9K':
        dl = {'Node': (0, 16), 'Type': (16, 42), 'State': (42, 59), 'Config State': (59, 77)}
    elif host.family == 'CRS':
        dl = {'Node': (0, 14), 'Type': (14, 32), 'State': (51, 67), 'Config State': (67, 82)}
    elif host.family == 'XR12K':
        dl = {'Node': (0, 16), 'Type': (16, 32), 'State': (47, 64), 'Config State': (64, 79)}
    else:
        ctx.warning("Unsupported platform {}".format(host.family))
        return None

    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:
        if line[0].isdigit():
            node = line[dl['Node'][0]:dl['Node'][1]].strip()
            if not re.search(r'CPU\d+$', node):
                continue

            node_type = line[dl['Type'][0]:dl['Type'][1]].strip()
            state = line[dl['State'][0]:dl['State'][1]].strip()
            config_state = line[dl['Config State'][0]:dl['Config State'][1]].strip()

            entry = {
                'type': node_type,
                'state': state,
                'config_state': config_state
            }
            inventory[node] = entry

    return inventory



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re
from csmpe.plugins import CSMPlugin
from utils import install_add_remove, install_add_ftp, install_add_scp
from csmpe.core_plugins.csm_get_inventory.ios_xe.plugin import get_package, get_inventory
from condoor.exceptions import CommandSyntaxError


class Plugin(CSMPlugin):
    """This plugin adds packages from repository to the device."""
    name = "Install Add Plugin"
    platforms = {'ASR900', 'ASR1K'}
    phases = {'Add'}
    os = {'XE'}

    def run(self):
        server_repository_url = self.ctx.server_repository_url

        if server_repository_url is None:
            self.ctx.error("No repository provided")
            return

        packages = self.ctx.software_packages
        if packages is None:
            self.ctx.error("No package list provided")
            return

        self.ctx.info("Add Package(s) Pending")
        self.ctx.post_status("Add Package(s) Pending")

        self.ctx.pause_session_logging()
        try:
            self.ctx.send('dir harddisk:')
            disk = 'harddisk:'
        except CommandSyntaxError:
            disk = 'bootflash:'
        self.ctx.resume_session_logging()

        for package in packages:

            stby_disk = ''
            self.ctx.pause_session_logging()
            output = self.ctx.send('dir ' + disk + package)
            self.ctx.resume_session_logging()

            m = re.search('No such file', output)

            if not m:
                self.ctx.info("No action: {} exists in {}".format(package, disk))
                continue

            if server_repository_url.startswith("tftp"):
                cmd = "copy {}/{} {}".format(server_repository_url, package, disk)
                install_add_remove(self.ctx, cmd)
            elif server_repository_url.startswith("ftp"):
                install_add_ftp(self.ctx, package, disk)
            elif server_repository_url.startswith("scp"):
                install_add_scp(self.ctx, package, disk)
            else:
                self.ctx.error("Unsupported repository type {}".format(server_repository_url))

            cmd = "dir " + 'stby-' + disk
            self.ctx.pause_session_logging()
            try:
                self.ctx.send(cmd)
                stby_disk = 'stby-' + disk
            except CommandSyntaxError:
                continue
            self.ctx.resume_session_logging()

            if stby_disk:
                cmd = "copy {}{} {}{}".format(disk, package, stby_disk, package)
                install_add_remove(self.ctx, cmd)

        self.ctx.info("Package(s) Added Successfully")

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================
import re
import time
import itertools
from condoor import ConnectionError, CommandError
from csmpe.core_plugins.csm_node_status_check.ios_xr.plugin_lib import parse_show_platform

install_error_pattern = re.compile(r"Error:    (.*)$", re.MULTILINE)

plugin_ctx = None


def send_yes(fsm_ctx):
    plugin_ctx.send('Y')
    return True


def log_install_errors(ctx, output):
    errors = re.findall(install_error_pattern, output)
    for line in errors:
        ctx.warning(line)


def watch_operation(ctx, op_id=0):
    """
    Function to keep watch on progress of operation
    and report KB downloaded.

    """
    no_install = r"There are no install requests in operation"
    op_progress = r"The operation is (\d+)% complete"
    op_download = r"(.*)KB downloaded: Download in progress"
    success = "Install operation {} completed successfully".format(op_id)

    cmd_show_install_request = "admin show install request"

    ctx.info("Watching the operation {} to complete".format(op_id))

    propeller = itertools.cycle(["|", "/", "-", "\\", "|", "/", "-", "\\"])

    output = None
    last_status = None
    finish = False
    time_tried = 0
    op_id = str(op_id)
    while not finish:
        try:
            try:
                # this is to catch the successful operation as soon as possible
                ctx.send("", wait_for_string=success, timeout=20)
                finish = True
            except ctx.CommandTimeoutError:
                pass

            message = ""
            # on CRS, it is observed that during Add, any command typed hangs for a while
            output = ctx.send(cmd_show_install_request, timeout=300)
            if op_id in output:
                # FIXME reconsider the logic here
                result = re.search(op_progress, output)
                if result:
                    status = result.group(0)
                    message = "{} {}".format(propeller.next(), status)

                result = re.search(op_download, output)
                if result:
                    status = result.group(0)
                    message += "\r\n<br>{}".format(status)

                if message != last_status:
                    ctx.post_status(message)
                    last_status = message
        except (ConnectionError, ctx.CommandTimeoutError) as e:
            if time_tried > 2:
                raise e

            time_tried += 1
            ctx.disconnect()
            time.sleep(60)
            ctx.reconnect()

        # ctx.send returns with an empty output from 'show install request'. go back to loop
        if output is None:
            continue

        # ctx.send returns and no active install action in progress. terminate the loop
        if no_install in output:
            break

    return output


def validate_node_state(inventory):
    valid_state = [
        'IOS XR RUN',
        'PRESENT',
        'READY',
        'FAILED',
        'OK',
        'DISABLED',
        'UNPOWERED',
        'ADMIN DOWN',
        'PWD',
        'NOT ALLOW ONLIN',  # This is not spelling error
    ]

    for key, value in inventory.items():
        if 'CPU' in key:
            if value['state'] not in valid_state:
                break
    else:
        return True
    return False


def wait_for_reload(ctx):
    """
     Wait for system to come up with max timeout as 25 Minutes

    """
    begin = time.time()
    if not ctx.is_console:
        # wait a little bit before disconnect so that newline character can reach the router
        # XR ddts CSCvb67386 workaround - reload is aborted after a confirmation is received from
        # the router if disconnect too soon (less than 6 seconds)
        time.sleep(10)
        ctx.disconnect()
        ctx.post_status("Waiting for device boot to reconnect")
        ctx.info("Waiting for device boot to reconnect")
        time.sleep(60)
        ctx.reconnect(max_timeout=3600, force_discovery=True)  # 60 * 60 = 3600
    else:
        ctx.info("Keeping console connected")
        ctx.post_status("Boot process started")
        ctx.info("Boot process started")
        if not ctx.reload(reload_timeout=1500, no_reload_cmd=True):
            ctx.error("Encountered error when attempting to reload device.")
        ctx.info("Boot process finished")

    ctx.info("Device connected successfully")

    timeout = 3600
    poll_time = 30
    time_waited = 0
    xr_run = "IOS XR RUN"

    cmd = "admin show platform"
    ctx.info("Waiting for all nodes to come up")
    ctx.post_status("Waiting for all nodes to come up")
    time.sleep(100)

    output = None

    while 1:
        # Wait till all nodes are in XR run state
        time_waited += poll_time
        if time_waited >= timeout:
            break

        time.sleep(poll_time)

        # show platform can take more than 1 minute after router reload. Issue No. 47
        output = ctx.send(cmd, timeout=600)
        if xr_run in output:
            inventory = parse_show_platform(ctx, output)
            if validate_node_state(inventory):
                ctx.info("All nodes in operational state")
                elapsed = time.time() - begin
                ctx.info("Overall outage time: {} minute(s) {:.0f} second(s)".format(elapsed // 60, elapsed % 60))
                return True

    # Some nodes did not come to run state
    ctx.error("Not all nodes have came up: {}".format(output))
    # this will never be executed
    return False


def watch_install(ctx, cmd, op_id=0):
    success_oper = r'Install operation (\d+) completed successfully'
    completed_with_failure = r'Install operation (\d+) completed with failure'
    failed_oper = r'Install operation (\d+) failed'
    failed_incr = r'incremental.*parallel'
    # restart = r'Parallel Process Restart'
    install_method = r'Install [M|m]ethod: (.*)'
    op_success = r"The install operation will continue asynchronously"

    watch_operation(ctx, op_id)

    output = ctx.send("admin show install log {} detail".format(op_id))
    if re.search(failed_oper, output):
        if re.search(failed_incr, output):
            ctx.info("Retrying with parallel reload option")
            cmd += " parallel-reload"
            output = ctx.send(cmd)
            if op_success in output:
                result = re.search(r'Install operation (\d+) \'', output)
                if result:
                    op_id = result.group(1)
                    watch_operation(ctx, op_id)
                    output = ctx.send("admin show install log {} detail".format(op_id))
                else:
                    log_install_errors(ctx, output)
                    ctx.error("Operation ID not found")
                    return
        else:
            log_install_errors(ctx, output)
            ctx.error(output)
            return

    """
    Command example:

    admin show install log 151 detail

Install operation 151 started by user 'root' via CLI at 04:42:17 PST Sun Sep 20 1987.
(admin) install activate disk0:asr9k-mini-px-6.1.3 disk0:asr9k-services-px-6.1.3 disk0:asr9k-doc-px-6.1.3 disk0:asr9k-video-px-6.1.3 disk0:asr9k-fpd-px-6.1.3 disk0:asr9k-mpls-px-6.1.3 disk0:asr9k-mcast-px-6.1.3 disk0:asr9k-px-5.3.3.CSCux41951-1.0.0 disk0:asr9k-mgbl-px-6.1.3 disk0:asr9k-optic-px-6.1.3 disk0:asr9k-k9sec-px-6.1.3 disk0:asr9k-9000v-nV-px-6.1.3 disk0:asr9k-bng-px-6.1.3 disk0:asr9k-li-px-6.1.3 disk0:asr9k-services-infra-6.1.3 disk0:asr9k-px-5.3.3.CSCux68183-1.0.0 prompt-level none
Install operation 151 completed successfully at 04:46:35 PST Sun Sep 20 1987.

Install logs:
    Install operation 151 '(admin) install activate disk0:asr9k-mini-px-6.1.3 disk0:asr9k-services-px-6.1.3 disk0:asr9k-doc-px-6.1.3 disk0:asr9k-video-px-6.1.3 disk0:asr9k-fpd-px-6.1.3 disk0:asr9k-mpls-px-6.1.3 disk0:asr9k-mcast-px-6.1.3 disk0:asr9k-px-5.3.3.CSCux41951-1.0.0 disk0:asr9k-mgbl-px-6.1.3 disk0:asr9k-optic-px-6.1.3 disk0:asr9k-k9sec-px-6.1.3 disk0:asr9k-9000v-nV-px-6.1.3 disk0:asr9k-bng-px-6.1.3 disk0:asr9k-li-px-6.1.3 disk0:asr9k-services-infra-6.1.3 disk0:asr9k-px-5.3.3.CSCux68183-1.0.0
    prompt-level none' started by user 'root' via CLI at 04:42:17 PST Sun Sep 20 1987.
    Info:     After this install operation, some SMU package(s) are fully/partially superceded. The fully superseded SMUs can found using CLI: 'show install superceded'. If found those can be deactivated using CLI: 'install deactivate superceded'.
    Warning:  There is no valid license for the following package:
    Warning:
    Warning:      disk0:asr9k-li-6.1.3
    Warning:
    Info:     The following sequence of sub-operations has been determined to minimize any impact:
    Info:
    Info:     Sub-operation 1:
    Info:         Install Method: Parallel Process Restart
    Info:         asr9k-px-5.3.3.CSCux41951-1.0.0
    Info:
    Info:     Sub-operation 2:
    Info:         Install Method: Parallel Process Restart
    Info:         asr9k-px-5.3.3.CSCux68183-1.0.0
    Info:
    Info:     Sub-operation 3:
    Info:         Install Method: Parallel Roeload
    Info:         asr9k-services-infra-6.1.3
    Info:         asr9k-li-px-6.1.3
    Info:         asr9k-bng-px-6.1.3
    Info:         asr9k-9000v-nV-px-6.1.3
    Info:         asr9k-k9sec-px-6.1.3
    Info:         asr9k-optic-px-6.1.3
    Info:         asr9k-mgbl-px-6.1.3
    Info:         asr9k-mcast-px-6.1.3
    Info:         asr9k-mpls-px-6.1.3
    Info:         asr9k-fpd-px-6.1.3
    Info:         asr9k-video-px-6.1.3
    Info:         asr9k-doc-px-6.1.3
    Info:         asr9k-services-px-6.1.3
    Info:         asr9k-mini-px-6.1.3
    Info:
    'prompt-level none' specified. Proceeding with operation.
    Info:     This operation will reload the following nodes in parallel:
    Info:         0/RSP0/CPU0 (RP) (SDR: Owner)
    Info:         0/RSP1/CPU0 (RP) (SDR: Owner)
    Info:         0/0/CPU0 (LC) (SDR: Owner)
    Info:         0/1/CPU0 (LC) (SDR: Owner)
    Info:         0/3/CPU0 (LC) (SDR: Owner)
    Info:         0/5/CPU0 (LC) (SDR: Owner)
    Info:         0/6/CPU0 (LC) (SDR: Owner)
    Install operation 151: load phase started at 04:43:32 PST Sun Sep 20 1987.
    Install operation 151: load phase started at 04:44:45 PST Sun Sep 20 1987.
    Install operation 151: load phase started at 04:45:59 PST Sun Sep 20 1987.
    Info:     The changes made to software configurations will not be persistent across system reloads. Use the command '(admin) install commit' to make changes persistent.
    Info:     Please verify that the system is consistent following the software change using the following commands:
    Info:         show system verify
    Info:         install verify packages
    Install operation 151 completed successfully at 04:46:35 PST Sun Sep 20 1987.

Summary:
    Sub-operation 1:
    Started at: 04:43:32 PST Sun Sep 20 1987
    Install method: Parallel Process Restart
    Summary of changes on nodes 0/RSP0/CPU0, 0/RSP1/CPU0:
        Activated:    asr9k-fwding-5.3.3.CSCux41951-1.0.0
        Impacted:     iosxr-mcast-5.3.3
        No processes affected

    Summary of changes on node 0/0/CPU0:
        Activated:    asr9k-fwding-5.3.3.CSCux41951-1.0.0
        Impacted:     asr9k-fwding-5.3.3
                      asr9k-li-5.3.3
            1 asr9k-li processes affected (0 updated, 0 added, 0 removed, 1 impacted)
            1 asr9k-fwding processes affected (0 updated, 0 added, 0 removed, 1 impacted)

    Summary of changes on node 0/1/CPU0:
        Activated:    asr9k-fwding-5.3.3.CSCux41951-1.0.0
        Impacted:     asr9k-fwding-5.3.3
                      asr9k-li-5.3.3
            1 asr9k-li processes affected (0 updated, 0 added, 0 removed, 1 impacted)
            1 asr9k-fwding processes affected (0 updated, 0 added, 0 removed, 1 impacted)

    Summary of changes on nodes 0/3/CPU0, 0/5/CPU0:
        Activated:    asr9k-fwding-5.3.3.CSCux41951-1.0.0
        Impacted:     asr9k-fwding-5.3.3
                      asr9k-li-5.3.3
            1 asr9k-li processes affected (0 updated, 0 added, 0 removed, 1 impacted)
            1 asr9k-fwding processes affected (0 updated, 0 added, 0 removed, 1 impacted)

    Summary of changes on node 0/6/CPU0:
        Activated:    asr9k-fwding-5.3.3.CSCux41951-1.0.0
        Impacted:     asr9k-fwding-5.3.3
                      asr9k-li-5.3.3
            1 asr9k-li processes affected (0 updated, 0 added, 0 removed, 1 impacted)
            1 asr9k-fwding processes affected (0 updated, 0 added, 0 removed, 1 impacted)

    Sub-operation 2:
    Started at: 04:44:45 PST Sun Sep 20 1987
    Install method: Parallel Process Restart
    Summary of changes on nodes 0/RSP0/CPU0, 0/RSP1/CPU0:
        Activated:    iosxr-ce-5.3.3.CSCux68183-1.0.0
            1 iosxr-ce-5.3.3.CSCux68183 processes affected (1 updated, 0 added, 0 removed, 0 impacted)

[snip]
    Sub-operation 3:
    Started at: 04:45:58 PST Sun Sep 20 1987
    Install method: Parallel Reoload
    Summary of changes on nodes 0/RSP0/CPU0, 0/RSP1/CPU0:
        Activated:    asr9K-doc-supp-6.1.3
                      asr9k-9000v-nV-supp-6.1.3
[snip]
    """
    # it produces a list of all captured groups matching the pattern, i.e.:
    # ['Parallel Process Restart', 'Parallel Process Restart', 'Parallel Reload', 'Parallel Process Restart' ....]
    result = re.findall(install_method, output)
    if result:
        if "Parallel Reload" in result:
            ctx.info("Parallel Reload Pending")
            if re.search(completed_with_failure, output):
                ctx.info("Install completed with failure, going for reload")
            elif re.search(success_oper, output):
                ctx.info("Install completed successfully, going for reload")
            return wait_for_reload(ctx)
        elif "Parallel Process Restart" in result:
            ctx.info("Parallel Process Restart Pending")
            ctx.info("Install completed successfully, going for process restart")
            return True
        else:
            ctx.warning("No Install Method detected.")

    log_install_errors(ctx, output)
    return False


def install_add_remove(ctx, cmd, has_tar=False):
    message = "Waiting the operation to continue asynchronously"
    ctx.info(message)
    ctx.post_status(message)

    output = ctx.send(cmd, timeout=7200)
    result = re.search(r'Install operation (\d+) \'', output)
    if result:
        op_id = result.group(1)
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation failed")
        return  # for sake of clarity

    op_success = "The install operation will continue asynchronously"
    failed_oper = r'Install operation {} failed'.format(op_id)
    if op_success in output:
        watch_operation(ctx, op_id=op_id)
        output = ctx.send("admin show install log {} detail".format(op_id))
        if re.search(failed_oper, output):
            log_install_errors(ctx, output)
            ctx.error("Operation {} failed".format(op_id))
            return  # for same of clarity

        ctx.info("Operation {} finished successfully".format(op_id))
        if has_tar is True:
            ctx.set_operation_id(ctx.software_packages, op_id)
            ctx.info("The operation {} stored".format(op_id))

        return  # for sake of clarity
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation {} failed".format(op_id))


def install_activate_deactivate(ctx, cmd):
    message = "Waiting the operation to continue asynchronously"
    ctx.info(message)
    ctx.post_status(message)

    op_success = "The install operation will continue asynchronously"
    output = ctx.send(cmd, timeout=7200)
    result = re.search(r'Install operation (\d+) \'', output)
    if result:
        op_id = result.group(1)
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation failed")
        return

    if op_success in output:
        success = watch_install(ctx, cmd, op_id)
        if not success:
            ctx.error("Reload or boot failure")
            return

        ctx.info("Operation {} finished successfully".format(op_id))
        return
    else:
        log_install_errors(ctx, output)
        ctx.error("Operation {} failed".format(op_id))
        return


def install_remove_all(ctx, cmd, hostname):
    """
    Success Condition:
    RP/0/RSP0/CPU0:RO#admin install remove inactive async

    Install operation 36 '(admin) install remove inactive' started by user 'root' via CLI at 01:14:44 PST Wed Feb 25 1987.
    / 1% complete: The operation can no longer be aborted (ctrl-c for options)
    Info:     This operation will remove the following package:
    Info:         disk0:asr9k-px-5.3.3.CSCuz14049-1.0.0
    - 1% complete: The operation can no longer be aborted (ctrl-c for options)
    \ 1% complete: The operation can no longer be aborted (ctrl-c for options)
    Info:     After this install remove the following install rollback point will no longer be reachable, as the required packages will not be present:
    Info:         15
    | 1% complete: The operation can no longer be aborted (ctrl-c for options)
    Proceed with removing these packages? [confirm]
    / 1% complete: The operation can no longer be aborted (ctrl-c for options)
    The install operation will continue asynchronously.
    RP/0/RSP0/CPU0:RO#Install operation 36 completed successfully at 01:14:51 PST Wed Feb 25 1987.

    Failed Conditions:
    RP/0/RSP0/CPU0:RO(admin)#install remove inactive
    Install operation 588 '(admin) install remove inactive' started by user 'lab'
    via CLI at 00:42:48 PST Tue Feb 24 1987.
    Error:    Cannot proceed with the remove operation because there are no
    Error:    packages that can be removed. Packages can only be removed if they
    Error:    are not part of the active software and not part of the committed
    Error:    software.
    Error:    Suggested steps to resolve this:
    Error:     - check the set of active packages using '(admin) show install
    Error:       active'.
    Error:     - check the committed software using '(admin) show install
    Error:       committed'.
    Error:     - check the set of inactive packages using '(admin) show install
    Error:       inactive'.
    Install operation 588 failed at 00:42:48 PST Tue Feb 24 1987.

    RP/0/RSP0/CPU0:RO(admin)#install remove inactive async
    Error:    Cannot proceed with the operation as another install command is
    Error:    currently in operation.
    Error:    Suggested steps to resolve this:
    Error:     - use 'show install request' to see the state of the current install
    Error:       operation.
    Error:     - re-issue the command when the current operation has completed.
    """  # noqa: W605

    global plugin_ctx
    plugin_ctx = ctx

    # no op_id is returned from XR for install remove inactive
    # need to figure out the last op_id first

    cmd_show_install_log_reverse = \
        'admin show install log reverse | utility egrep "Install operation [0-9]+ started"'
    output = ctx.send(cmd_show_install_log_reverse, timeout=300)

    if 'No log information' in output:
        op_id = 0
    else:
        result = re.search(r'Install operation (\d+) started', output)
        if result:
            op_id = int(result.group(1))
        else:
            log_install_errors(ctx, output)
            ctx.error("Operation ID not found by admin show install log reverse")
            return

    # Expected Operation ID
    op_id += 1

    operr = "Install operation {} failed at".format(op_id)
    Error1 = re.compile(r"Error:     - re-issue the command when the current operation has completed.")
    Error2 = re.compile(operr)
    Proceed_removing = re.compile(r"\[confirm\]")
    Host_prompt = re.compile(hostname)

    events = [Host_prompt, Error1, Error2, Proceed_removing]
    transitions = [
        (Error1, [0], -1, CommandError("Another install command is currently in operation", hostname), 1800),
        (Error2, [0], -1, CommandError("No packages can be removed", hostname), 1800),
        (Proceed_removing, [0], 2, send_yes, 1800),
        (Host_prompt, [2], -1, None, 1800),
    ]

    if not ctx.run_fsm("Remove Inactive All", cmd, events, transitions, timeout=1800):
        ctx.error("Failed: {}".format(cmd))

    message = "Waiting the operation to continue asynchronously"
    ctx.info(message)
    ctx.post_status(message)

    last_status = None
    no_install = r"There are no install requests in operation"
    op_progress = r"The operation is (\d+)% complete"
    cmd_show_install_request = "admin show install request"
    op_success = "Install operation {} completed successfully".format(op_id)
    propeller = itertools.cycle(["|", "/", "-", "\\", "|", "/", "-", "\\"])

    finish = False
    time_tried = 0
    op_id = str(op_id)
    while not finish:
        try:
            try:
                # this is to catch the successful operation as soon as possible
                ctx.send("", wait_for_string=op_success, timeout=20)
                finish = True
            except ctx.CommandTimeoutError:
                pass

            message = ""
            # on CRS, it is observed that during Add, any command typed hangs for a while
            output = ctx.send(cmd_show_install_request, timeout=300)
            if op_id in output:
                result = re.search(op_progress, output)
                if result:
                    status = result.group(0)
                    message = "{} {}".format(propeller.next(), status)

                if message != last_status:
                    ctx.post_status(message)
                    last_status = message
        except (ConnectionError, ctx.CommandTimeoutError) as e:
            if time_tried > 120:
                raise e

            time_tried += 1
            time.sleep(30)

        if no_install in output:
            break

    cmd_show_install_log = "admin show install log {} detail".format(op_id)
    output = ctx.send(cmd_show_install_log, timeout=300)
    ctx.info(output)

    if op_success in output:
        message = "Remove All Inactive Package(s) Successfully"
        ctx.info(message)
        ctx.post_status(message)
    else:
        ctx.error("Remove All Inactive Package(s) failed")


def build_satellite_list(satellite_ids):
    """
    :param satellite_ids: a string of satellite IDs with comma delimiter and range
    For example, satellite_ids = '100-102,105,106-109,110,160-164,319-320'
    :return: a list of satellite IDs in string
    """

    L = []
    input_list = satellite_ids.split(',')
    for item in input_list:
        if '-' in item:
            m = re.search(r'(\w+)-(\w+)', item)
            arg1 = int(m.group(1))
            arg2 = int(m.group(2)) + 1
            ilist = range(arg1, arg2)
            slist = []
            for i in ilist:
                slist.append(str(i))
            L = L + slist
        else:
            L.append(item)

    return L


def build_new_argument(L):
    """
    :param L: a list of satellite id in string
    :return: a string of satellite id that includes comma delimiter and range
    """

    LI = [int(x) for x in L]
    LI.sort()

    lrange = len(LI) - 1

    if lrange == 0:
        arg = str(LI[0])
    else:
        # walk thru the list to construct the new string argument
        arg = str(LI[0])
        id = 1
        nexts = ''
        while id < lrange:
            if LI[id - 1] + 1 == LI[id]:
                if nexts != '-':
                    nexts = '-'
            else:
                if nexts == '-':
                    arg = arg + '-' + str(LI[id - 1]) + ',' + str(LI[id])
                    nexts = ''
                else:
                    arg = arg + ',' + str(LI[id])
            id += 1

        id = lrange
        if LI[id - 1] + 1 == LI[id]:
            if nexts == '-':
                arg = arg + '-' + str(LI[id])
            else:
                arg = arg + ',' + str(LI[id])
        else:
            if nexts == '-':
                arg = arg + '-' + str(LI[id - 1]) + ',' + str(LI[id])
                nexts = ''
            else:
                arg = arg + ',' + str(LI[id])

    return arg


def install_satellite_transfer(ctx, satellite_ids):
    """
    RP/0/RP0/CPU0:AGN_PE_11_9k#install nv satellite 160,163 transfer
    Install Op 2: install nv satellite 160,163 transfer
        2 configured satellites have been specified for the transfer operation.
        2 satellites have successfully initiated the transfer operation.

    RP/0/RP0/CPU0:AGN_PE_11_9k#show nv satellite status satellite 161
    Satellite 161
    -------------
      Status: Connected (Transferring new image)
      Redundancy: Active (Group: 100)
      Type: ncs5002
      Displayed device name: Sat161
      MAC address: c472.95a6.87a5
      IPv4 address: 10.0.161.1 (auto, VRF: **nVSatellite)
      Serial Number: FOC1928R0XP
      Remote version: Compatible (older version)
        IOFPGA: 0.17
        MB_MIFPGA: 0.16
        DB_MIFPGA: 0.16
        BIOS: 1.11
        XR: 6.2.25.03I (Available: 6.2.25.05I)
      Received candidate fabric ports:
        nVFabric-TenGigE0/0/78-79 (permanent)
        nVFabric-HundredGigE0/1/0-3 (permanent)
      Configured satellite fabric links:
        HundredGigE0/18/0/5
        -------------------
          Status: Satellite Ready
          Remote ports: TenGigE0/0/0-70

    RP/0/RP0/CPU0:AGN_PE_11_9k#show nv satellite status satellite 161
    Satellite 161
    -------------
      Status: Connected (New image transferred)
      Redundancy: Active (Group: 100)
      Type: ncs5002
      Displayed device name: Sat161
      MAC address: c472.95a6.87a5
      IPv4 address: 10.0.161.1 (auto, VRF: **nVSatellite)
      Serial Number: FOC1928R0XP
      Remote version: Compatible (older version)
        IOFPGA: 0.17
        MB_MIFPGA: 0.16
        DB_MIFPGA: 0.16
        BIOS: 1.11
        XR: 6.2.25.03I (Available: 6.2.25.05I)
      Received candidate fabric ports:
        nVFabric-TenGigE0/0/78-79 (permanent)
        nVFabric-HundredGigE0/1/0-3 (permanent)
      Configured satellite fabric links:
        HundredGigE0/18/0/5
        -------------------
          Status: Satellite Ready
          Remote ports: TenGigE0/0/0-70
    """

    global plugin_ctx
    plugin_ctx = ctx

    # satellite_ids = '100-102,105,106-109,110,160-164,319-320'
    L = build_satellite_list(satellite_ids)
    ctx.info("Checking satellite status: {}".format(','.join(L)))

    # filter invalid cases
    command_success = True
    for id in range(len(L)):
        show_cmd = 'show nv satellite status satellite ' + L[id]
        output = ctx.send(show_cmd)

        if 'No information for satellite' in output:
            ctx.warning("There is no information for Satellite ID {}.".format(L[id]))
            ctx.warning("{}".format(output))
            L[id] = None
            command_success = False

        if 'Status: Connected' not in output:
            ctx.warning("Satellite ID {} Status is not Connected.".format(L[id]))
            ctx.warning("{}".format(output))
            L[id] = None
            command_success = False

        if 'Status: Connected (New image transferred)' in output:
            ctx.info("Satellite ID {} Status: Connected (New image transferred)".format(L[id]))
            L[id] = None

        # Replace 'Remote version: Compatible (latest version)' in output with
        # 'Available' not in output to workaround XR 5.1.3 bug that even when there is a
        # new satellite software available it will display Compatible (latest version).
        # if 'Remote version: Compatible (latest version)' in output:
        if 'Available' not in output:
            ctx.info("Satellite ID {} Remote version: Compatible (latest version)".format(L[id]))
            L[id] = None

    L = [x for x in L if x is not None]

    if not L:
        if command_success:
            ctx.info("Satellite-Transfer: all satellites are up to date. No action is required.")
            return True
        else:
            ctx.warning("Satellite-Transfer: one or more satellites are not ready")
            return False

    # construct the new argument
    ctx.info("Satellite-Transfer satellites {}".format(','.join(L)))
    arg = build_new_argument(L)

    cmd = 'install nv satellite ' + arg + ' transfer'
    Warning = re.compile(r"Do you wish to continue\? \[confirm\(y/n\)\]")
    Host_prompt = re.compile(ctx._connection.hostname)

    events = [Host_prompt, Warning]
    transitions = [
        (Warning, [0], -1, send_yes, 30),
        (Host_prompt, [0], -1, None, 30),
    ]

    if not ctx.run_fsm("Satellite-Transfer ", cmd, events, transitions, timeout=30):
        ctx.warning("Failed: {}".format(cmd))
        return False

    ctx.info("Waiting for Satellite-Transfer to complete")
    ctx.post_status("Waiting for Satellite-Transfer to complete")

    timeout = 3600
    poll_time = 180
    time_waited = 3
    begin = time.time()
    time.sleep(time_waited)

    header = "Transferring Satellite Image <br>" + \
             "<pre>" + \
             "Sat-ID  Type       Status\n" + \
             "------  --------   ----------------\n"

    while 1:
        # Waiting for transfer to complete for all satellites
        if time_waited >= timeout:
            break

        status = header
        show_cmd = 'show nv satellite status brief'
        output = ctx.send(show_cmd)
        lines = output.split('\n')
        lines = [x for x in lines if x]

        for line in lines:

            if not line[0].isdigit():
                continue

            sl = line.split()
            elements = len(sl)
            if elements > 5:
                i = 5
                while i < elements:
                    sl[4] = sl[4] + ' ' + sl[i]
                    i += 1

            for id in range(len(L)):
                if L[id] == sl[0]:
                    if 'Transferred' in sl[4]:
                        L[id] = None
                    else:
                        length = len(sl[0])
                        space = 8 - length
                        sat_id = sl[0] + ' ' * space
                        length = len(sl[1])
                        space = 11 - length
                        type = sl[1] + ' ' * space
                        status = status + sat_id + type + sl[4] + '\n'

                    break

        L = [x for x in L if x is not None]

        if not L:
            elapsed = time.time() - begin
            ctx.info("Satellite-Transfer time: {} minute(s) {:.0f} "
                     "second(s)".format(elapsed // 60, elapsed % 60))
            if command_success:
                ctx.info("Satellite-Transfer completed for all the satellites")
                ctx.post_status("Satellite-Transfer completed for all the satellites")
                return True
            else:
                ctx.warning("Satellite-Transfer completed but some satellites were not ready.")
                ctx.post_status("Satellite-Transfer completed but some satellites were not ready.")
                return True
        else:
            status += "</pre>"
            ctx.post_status(status)
            time_waited += poll_time
            time.sleep(poll_time)

    # Some transfer did not ccomplete
    ctx.warning("Satellite-Transfer did not complete for satellites {}.:".format(','.join(L)))
    return False


def install_satellite_activate(ctx, satellite_ids):
    """
    RP/0/RP0/CPU0:AGN_PE_11_9k#install nv satellite 160,161 activate
    The operation will cause an image to be transferred where required, and then activate new versions on the \
    requested satellites.
    WARNING: This may take the requested satellites out of service.
    Do you wish to continue? [confirm(y/n)] y
    Install Op 4: install nv satellite 160-161 activate
      2 configured satellites have been specified for the activate operation.
      2 satellites have successfully initiated the activate operation.

    RP/0/RP0/CPU0:AGN_PE_11_9k#show nv satellite status satellite 161
    Satellite 161
    -------------
      Status: Connected (Installing new image)
      Redundancy: Active (Group: 100)
      Type: ncs5002
      Displayed device name: Sat161
      MAC address: c472.95a6.87a5
      IPv4 address: 10.0.161.1 (auto, VRF: **nVSatellite)
      Serial Number: FOC1928R0XP
      Remote version: Compatible (older version)
        IOFPGA: 0.17
        MB_MIFPGA: 0.16
        DB_MIFPGA: 0.16
        BIOS: 1.11
        XR: 6.2.25.03I (Available: 6.2.25.05I)
      Received candidate fabric ports:
        nVFabric-TenGigE0/0/78-79 (permanent)
        nVFabric-HundredGigE0/1/0-3 (permanent)
      Configured satellite fabric links:
        HundredGigE0/18/0/5
        -------------------
          Status: Satellite Ready
          Remote ports: TenGigE0/0/0-70

    RP/0/RP0/CPU0:AGN_PE_11_9k#show nv satellite status satellite 161
    Satellite 161
    -------------
      Status: Discovery Stalled; Conflict: interface is down
      Type: ncs5002
      Displayed device name: Sat161
      IPv4 address: 10.0.161.1 (auto, VRF: **nVSatellite)
      Serial Number: FOC1928R0XP
      Configured satellite fabric links:
        HundredGigE0/18/0/5
        -------------------
          Status: Discovery Stalled; Conflict: interface is down
          Remote ports: TenGigE0/0/0-70

    RP/0/RP0/CPU0:AGN_PE_11_9k#show nv satellite status satellite 161
    Satellite 161
    -------------
      Status: Connected (Stable)
      Redundancy: Active (Group: 100) (Recovery Delay remaining: 4m 23s)
      Type: ncs5002
      Displayed device name: Sat161
      MAC address: c472.95a6.87a5
      IPv4 address: 10.0.161.1 (auto, VRF: **nVSatellite)
      Serial Number: FOC1928R0XP
      Remote version: Compatible (latest version)
        IOFPGA: 0.17
        MB_MIFPGA: 0.16
        DB_MIFPGA: 0.16
        BIOS: 1.11
        XR: 6.2.25.05I (Latest)
      Received candidate fabric ports:
        nVFabric-TenGigE0/0/78-79 (permanent)
        nVFabric-HundredGigE0/1/0-3 (permanent)
      Configured satellite fabric links:
        HundredGigE0/18/0/5
        -------------------
          Status: Satellite Ready
          Remote ports: TenGigE0/0/0-70
    """

    global plugin_ctx
    plugin_ctx = ctx

    # satellite_ids = '100-102,105,106-109,110,160-164,319-320'
    L = build_satellite_list(satellite_ids)
    ctx.info("Checking satellite status: {}".format(','.join(L)))

    # filter invalid cases
    command_success = True
    for id in range(len(L)):
        show_cmd = 'show nv satellite status satellite ' + L[id]
        output = ctx.send(show_cmd)

        if 'No information for satellite' in output:
            ctx.warning("There is no information for Satellite ID {}.".format(L[id]))
            ctx.warning("{}".format(output))
            L[id] = None
            command_success = False

        if 'Status: Connected' not in output:
            ctx.warning("Satellite ID {} Status is not Connected.".format(L[id]))
            ctx.warning("{}".format(output))
            L[id] = None
            command_success = False

        # Replace 'Remote version: Compatible (latest version)' in output with
        # 'Available' not in output to workaround XR 5.1.3 bug that even when there is a
        # new satellite software available it will display Compatible (latest version).
        # if 'Remote version: Compatible (latest version)' in output:
        if 'Available' not in output:
            ctx.info("Satellite ID {} Remote version: Compatible (latest version)".format(L[id]))
            L[id] = None

    L = [x for x in L if x is not None]

    if not L:
        if command_success:
            ctx.info("Satellite-Activate: all satellites are up to date. No action is required.")
            return True
        else:
            ctx.warning("Satellite-Activate: one or more satellites are not ready")
            return False

    # construct the new argument
    ctx.info("Satellite-Activate satellites {}".format(','.join(L)))
    arg = build_new_argument(L)

    cmd = 'install nv satellite ' + arg + ' activate'
    Warning = re.compile(r"Do you wish to continue\? \[confirm\(y/n\)\]")
    Host_prompt = re.compile(ctx._connection.hostname)

    events = [Host_prompt, Warning]
    transitions = [
        (Warning, [0], -1, send_yes, 30),
        (Host_prompt, [0], -1, None, 30),
    ]

    if not ctx.run_fsm("Satellite-Activate", cmd, events, transitions, timeout=30):
        ctx.warning("Failed: {}".format(cmd))
        return False

    ctx.info("Waiting for Satellite-Activate to complete")
    ctx.post_status("Waiting for Satellite-Activate to complete")

    timeout = 5400
    poll_time = 180
    time_waited = 3
    begin = time.time()
    time.sleep(time_waited)

    header = "Activating Satellite Image <br>" + \
             "<pre>" + \
             "Sat-ID  Type       Status\n" + \
             "------  --------   ----------------\n"

    while 1:
        # Waiting for transfer to complete for all satellites
        if time_waited >= timeout:
            break

        status = header
        show_cmd = 'show nv satellite status brief'
        output = ctx.send(show_cmd)
        lines = output.split('\n')
        lines = [x for x in lines if x]

        for line in lines:

            if not line[0].isdigit():
                continue

            sl = line.split()
            elements = len(sl)
            if elements > 5:
                i = 5
                while i < elements:
                    sl[4] = sl[4] + ' ' + sl[i]
                    i += 1

            for id in range(len(L)):
                if L[id] == sl[0]:
                    if sl[4] == 'Connected' or sl[4] == 'Connected (Act)' or sl[4] == 'Connected (Stby)':
                        L[id] = None
                    else:
                        length = len(sl[0])
                        space = 8 - length
                        sat_id = sl[0] + ' ' * space
                        length = len(sl[1])
                        space = 11 - length
                        type = sl[1] + ' ' * space
                        status = status + sat_id + type + sl[4] + '\n'

                    break

        L = [x for x in L if x is not None]

        if not L:
            elapsed = time.time() - begin
            ctx.info("Satellite-Activate time: {} minute(s) {:.0f} "
                     "second(s)".format(elapsed // 60, elapsed % 60))
            if command_success:
                ctx.info("Satellite-Activate completed for all the satellites")
                ctx.post_status("Satellite-Activate completed for all the satellites")
                return True
            else:
                ctx.warning("Satellite-Activate completed but some satellites were not ready.")
                ctx.post_status("Satellite-Activate completed but some satellites were not ready.")
                return True
        else:
            status += "</pre>"
            ctx.post_status(status)
            time_waited += poll_time
            time.sleep(poll_time)

    # Some transfer did not ccomplete
    ctx.warning("Satellite-Activate did not complete for satellites {}.".format(','.join(L)))
    return False


def parse_pkg_list(output):
    """
    :param output: show install active/inactive summary
    :return: a list of active/inactive packages
    """

    pkg_list = []
    flag = False

    lines = output.split('\n')
    lines = [x for x in lines if x]

    for line in lines:
        if 'ctive Packages:' in line:
            flag = True
            continue

        if flag:
            if line[2].isalnum():
                break
            else:
                pkg = line.strip()
                pkg_list.append(pkg)

    return pkg_list


def report_changed_pkg(p, q):
    """
    :param p: a list of packages
    :param q: a list of packages
    :return: a list of packages in q but not in p
    """

    changed = list(set(q) - set(p))
    changed_list = [re.sub(r'^.*:', '', pkg) for pkg in changed]
    changed_list.sort()

    return changed_list



# =============================================================================
# migrate_system.py - plugin for migrating classic XR to eXR/fleXR
#
# Copyright (c)  2013, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re
import time

from csmpe.plugins import CSMPlugin
from migration_lib import check_exr_final_band, log_and_post_status, run_additional_custom_commands
from csmpe.core_plugins.csm_get_inventory.exr.plugin import get_package, get_inventory

TIMEOUT_FOR_COPY_CONFIG = 3600


class Plugin(CSMPlugin):
    """
    A plugin for loading configurations and upgrading FPD's
    after the system migrated to ASR9K IOS-XR 64 bit(eXR).
    If any FPD needs reloads after upgrade, the device
    will be reloaded after the upgrade.
    Console access is needed.
    """
    name = "Post-Migrate Plugin"
    platforms = {'ASR9K'}
    phases = {'Post-Migrate'}

    def _check_fpds_for_upgrade(self):
        """Check if any FPD's need upgrade, if so, upgrade all FPD's on all locations."""

        self.ctx.send("admin")

        fpdtable = self.ctx.send("show hw-module fpd")

        match = re.search(r"\d+/\w+.+\d+.\d+\s+[-\w]+\s+(NEED UPGD)", fpdtable)

        if match:
            if re.search(r"\d+/\w+.+\d+.\d+\s+[-\w]+\s+(NOT READY)", fpdtable):
                self.ctx.send("exit")
                self.ctx.error("Some nodes are not ready for FPD upgrade. Please re-schedule Post-Migrate when all nodes are ready.")
            else:
                return self._upgrade_all_fpds(len(re.findall("NEED UPGD|CURRENT", fpdtable)))

        self.ctx.send("exit")
        return True

    def _upgrade_all_fpds(self, num_fpds):
        """
        Upgrade all FPD's on all locations.
        If after all upgrade completes, some show that a reload is required to reflect the changes,
        the device will be reloaded.

        :param num_fpds: the number of FPD's that are in CURRENT and NEED UPGD states before upgrade.
        :return: True if upgraded successfully and reloaded(if necessary).
                 False if some FPD's did not upgrade successfully in 9600 seconds.
        """
        log_and_post_status(self.ctx, "Upgrading all FPD's.")
        self.ctx.send("upgrade hw-module location all fpd all")

        timeout = 9600
        poll_time = 30
        time_waited = 0

        time.sleep(60)
        while 1:
            # Wait till all FPDs finish upgrade
            time_waited += poll_time
            if time_waited >= timeout:
                break
            time.sleep(poll_time)
            output = self.ctx.send("show hw-module fpd")
            num_need_reload = len(re.findall("RLOAD REQ", output))
            if len(re.findall("CURRENT|UPGD SKIP", output)) + num_need_reload >= num_fpds:
                if num_need_reload > 0:
                    log_and_post_status(self.ctx,
                                        "Finished upgrading FPD(s). Reloading device to complete the upgrade.")
                    self.ctx.send("exit")
                    return self._reload_all()
                self.ctx.send("exit")
                return True

        output = self.ctx.send("show hw-module fpd")
        if len(re.findall(r"\d+% UPGD|IN QUEUE|NEED UPGD", output)) == 0:
            if len(re.findall("RLOAD REQ", output)) > 0:
                log_and_post_status(self.ctx, "Reloading device to complete the upgrade.")
                self.ctx.send("exit")
                return self._reload_all()
        else:
            self.ctx.warning(self.ctx, "FPD Upgrade not completed after {} minutes.".format(timeout / 60))

        self.ctx.send("exit")
        return True

    def _reload_all(self):
        """Reload the device with 1 hour maximum timeout"""
        if self.ctx.is_console:
            if not self.ctx.reload(reload_timeout=3600):
                self.ctx.error("Encountered error when attempting to reload device.")
        else:
            def send_yes(fsm_ctx):
                fsm_ctx.ctrl.sendline('yes')
                return True

            CONFIRM = re.compile(r"Reload hardware module \? \[no,yes\]")

            events = [CONFIRM]
            transitions = [
                (CONFIRM, [0], -1, send_yes, 0),
            ]

            if not self.ctx.run_fsm("RELOAD", "admin hw-module location all reload", events, transitions, timeout=300):
                self.ctx.error("Failed to reload.")

            # wait a little bit before disconnect so that newline character can reach the router
            time.sleep(5)
            self.ctx.disconnect()
            self.ctx.post_status("Waiting for device boot to reconnect")
            self.ctx.info("Waiting for device boot to reconnect")
            time.sleep(180)
            self.ctx.reconnect(max_timeout=3600, force_discovery=True)  # 60 * 60 = 3600

        return check_exr_final_band(self.ctx)

    def run(self):

        if self.ctx.os_type != "eXR":
            self.ctx.error("Device is not running ASR9K-X64.")

        check_exr_final_band(self.ctx, timeout=1500)

        log_and_post_status(self.ctx, "Capturing new IOS XR and Calvados configurations.")

        run_additional_custom_commands(self.ctx, {"show running-config", "admin show running-config"})

        self._check_fpds_for_upgrade()

        run_additional_custom_commands(self.ctx, {"show platform"})

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)

        return True



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from package_lib import SoftwarePackage
from csmpe.plugins import CSMPlugin
from install import install_add_remove, parse_pkg_list, report_changed_pkg
from csmpe.core_plugins.csm_get_inventory.ios_xr.plugin import get_package, get_inventory


class Plugin(CSMPlugin):
    """This plugin removes inactive packages from the device."""
    name = "Install Remove Plugin"
    platforms = {'ASR9K', 'XR12K', 'CRS'}
    phases = {'Remove'}
    os = {'XR'}

    def run(self):
        packages = self.ctx.software_packages
        if packages is None:
            self.ctx.error("No package list provided")
            return

        # If a wide card character is detected, like 'disk0:*6.0.2*', only the first element
        # in the list is considered valid.  An example CLI is 'admin install remove disk0:*6.0.2*'.
        if any('*' in package for package in packages):
            to_remove = packages[0]
        else:
            pkgs = SoftwarePackage.from_package_list(packages)

            installed_inact = SoftwarePackage.from_show_cmd(self.ctx.send("admin show install inactive summary"))
            packages_to_remove = pkgs & installed_inact

            if not packages_to_remove:
                self.ctx.warning("Packages already removed. Nothing to be removed")
                return

            to_remove = " ".join(map(str, packages_to_remove))

        cmd = 'admin install remove {} prompt-level none async'.format(to_remove)

        output = self.ctx.send("show install inactive summary")
        before = parse_pkg_list(output)

        self.ctx.info("Remove Package(s) Pending")
        self.ctx.post_status("Remove Package(s) Pending")

        install_add_remove(self.ctx, cmd)

        self.ctx.info("Package(s) Removed Successfully")

        output = self.ctx.send("show install inactive summary")
        after = parse_pkg_list(output)

        pkg_list = report_changed_pkg(after, before)
        pkg_change_list = ','.join(pkg_list)

        self.ctx.save_job_data("package_change_list", pkg_change_list)
        self.ctx.info("package change list:")

        for pkg in pkg_list:
            self.ctx.info("\t{}".format(pkg))

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from csmpe.plugins import CSMPlugin
from install import observe_install_remove_all
from csmpe.core_plugins.csm_get_inventory.exr.plugin import get_package, get_inventory


class Plugin(CSMPlugin):
    """This plugin removes inactive packages from the device."""
    name = "Install Remove All Inactive Plugin"
    platforms = {'ASR9K', 'NCS1K', 'NCS1001', 'NCS4K', 'NCS5K', 'NCS540',
                 'NCS5500', 'NCS6K', 'IOSXRv-9K', 'IOSXRv-X64'}
    phases = {'Remove All Inactive'}
    os = {'eXR'}

    def run(self):

        cmd = 'install remove inactive all'

        self.ctx.info("Remove All Inactive Package(s) Pending")
        self.ctx.post_status("Remove All Inactive Package(s) Pending")

        observe_install_remove_all(self.ctx, cmd, self.ctx.prompt)

        self.ctx.info("Remove All Inactive Package(s) Successfully")

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import time
from csmpe.plugins import CSMPlugin
from fpd_upgd_lib import fpd_is_current, fpd_needs_upgd, fpd_check_status, \
    hw_fpd_upgd, hw_fpd_reload, wait_for_fpd_upgd
from install import wait_for_reload
from csmpe.core_plugins.csm_get_inventory.exr.plugin import get_package, get_inventory
from csmpe.core_plugins.csm_install_operations.utils import update_device_info_udi


class Plugin(CSMPlugin):
    """This plugin removes inactive packages from the device."""
    name = "Install FPD Upgrade Plugin"
    platforms = {'ASR9K', 'NCS1K', 'NCS1001', 'NCS4K', 'NCS5K', 'NCS540',
                 'NCS5500', 'NCS6K', 'IOSXRv-9K', 'IOSXRv-X64'}
    phases = {'FPD-Upgrade'}
    os = {'eXR'}

    def run(self):

        self.ctx.info("FPD-Upgrade Pending")
        self.ctx.post_status("FPD-Upgrade Pending")

        fpd_location = self.ctx.load_job_data('fpd_location')[0]
        fpd_type = self.ctx.load_job_data('fpd_type')[0]

        self.ctx.info("fpd_location = {}".format(fpd_location))
        self.ctx.info("fpd_type = {}".format(fpd_type))

        # case 1: both fpd_location and fpd_type are none.
        if not fpd_location and not fpd_type:
            fpd_location = 'all'
            fpd_type = 'all'
        # case 2: only fpd_location is specified
        # case 3: only fpd_type is specified
        # case 4: both fpd_location and fpd_type are specified

        if fpd_is_current(self.ctx, fpd_location, fpd_type):
            self.ctx.info("All FPD devices are CURRENT. Nothing to be upgraded.")
            return True

        upgd_locations = fpd_needs_upgd(self.ctx, fpd_location, fpd_type)
        self.ctx.info("locations to be upgraded = {}".format(upgd_locations))
        if upgd_locations:
            if fpd_location == 'all' and fpd_type == 'all':
                if not hw_fpd_upgd(self.ctx, fpd_location, fpd_type):
                    self.ctx.error("Fail to issue {}".format('upgrade hw-module location all fpd all'))
                    return
                wait_for_fpd_upgd(self.ctx, fpd_location, fpd_type)
            elif fpd_location and not fpd_type:
                type = 'all'
                if not hw_fpd_upgd(self.ctx, fpd_location, type):
                    cmd = 'upgrade hw-module location ' + fpd_location + ' fpd ' + type
                    self.ctx.error("Fail to issue {}".format(cmd))
                    return
                wait_for_fpd_upgd(self.ctx, fpd_location, type)
            elif not fpd_location and fpd_type:
                for location in upgd_locations:
                    if not hw_fpd_upgd(self.ctx, location, fpd_type):
                        cmd = 'upgrade hw-module location ' + location + ' fpd ' + fpd_type
                        self.ctx.error("Fail to issue {}".format(cmd))
                        return
                    wait_for_fpd_upgd(self.ctx, location, fpd_type)
                    time.sleep(30)  # CLI may not work if issuing too quickly
            else:
                if not hw_fpd_upgd(self.ctx, fpd_location, fpd_type):
                    cmd = 'upgrade hw-module location ' + fpd_location + ' fpd ' + fpd_type
                    self.ctx.error("Fail to issue {}".format(cmd))
                    return
                wait_for_fpd_upgd(self.ctx, fpd_location, fpd_type)

        if not fpd_location and fpd_type:
            # check if RP0 / RP1 is to be reloaded
            location = None
            if upgd_locations:
                for location in upgd_locations:
                    if 'RP' in location or 'RSP' in location:
                        location = 'all'
                        break

            if location == 'all':
                if not hw_fpd_reload(self.ctx, location):
                    cmd = 'admin hw-module location ' + location + ' reload'
                    self.ctx.error("Fail to issue {}".format(cmd))
                    return
            else:
                if upgd_locations:
                    for location in upgd_locations:
                        if not hw_fpd_reload(self.ctx, location):
                            cmd = 'admin hw-module location ' + location + ' reload'
                            self.ctx.error("Fail to issue {}".format(cmd))
                            return
                        time.sleep(10)
        else:
            if not hw_fpd_reload(self.ctx, fpd_location):
                cmd = 'admin hw-module location ' + fpd_location + ' reload'
                self.ctx.error("Fail to issue {}".format(cmd))
                return

        success = wait_for_reload(self.ctx)
        if not success:
            self.ctx.error("Reload or boot failure")
            return

        self.ctx.info("Refreshing package and inventory information")
        self.ctx.post_status("Refreshing package and inventory information")
        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)

        update_device_info_udi(self.ctx)

        if fpd_check_status(self.ctx, fpd_location, fpd_type):
            self.ctx.info("FPD-Upgrade Successfully")
            return True
        else:
            self.ctx.error("FPD-Upgrade Completed but the status of one or more nodes is not Current or N/A")
            return False



"""This is a module providing FSM actions for Install Operations plugins."""
import re


def a_error(plugin_ctx, ctx):
    """Display the error message."""
    message = ctx.ctrl.after.strip().splitlines()[-1]
    plugin_ctx.error(message)
    return False


def a_remote_host_address(plugin_ctx, fsm_ctx):
    """send remote host address or name."""
    address = plugin_ctx.src_address
    fsm_ctx.ctrl.sendline(address)
    return True


def a_source_username(plugin_ctx, fsm_ctx):
    """send source username."""
    src_username = plugin_ctx.src_username
    fsm_ctx.ctrl.sendline(src_username)
    return True


def a_source_password(plugin_ctx, fsm_ctx):
    """send source password."""
    src_password = plugin_ctx.src_password
    fsm_ctx.ctrl.send_command(src_password, password=True)
    return True


def a_source_filename(plugin_ctx, fsm_ctx):
    """send source filename."""
    src_file = plugin_ctx.src_file
    fsm_ctx.ctrl.sendline(src_file)
    return True


def a_destination(plugin_ctx, fsm_ctx):
    """send destination filename."""
    destination_file = plugin_ctx.destination_file
    fsm_ctx.ctrl.sendline(destination_file)
    return True


def a_send_newline(fsm_ctx):
    """Display the error message."""
    fsm_ctx.sendline('\r\n')
    return True


def a_bytes_copied(plugin_ctx, fsm_ctx):
    """Capture number of bytes copied."""
    m = re.search(r"\d+ bytes copied in .* secs", fsm_ctx.ctrl.before)
    if m:
        plugin_ctx.info('{}'.format(m.group(0)))
    else:
        plugin_ctx.info('{}'.format(fsm_ctx.ctrl.before))

    return True



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import subprocess
from csmpe.plugins import CSMPlugin
from install import install_satellite_activate
from csmpe.core_plugins.csm_get_inventory.ios_xr.plugin import get_satellite


class Plugin(CSMPlugin):
    """This plugin removes all inactive packages from the device."""
    name = "Satellite-Activate Plugin"
    platforms = {'ASR9K'}
    phases = {'Satellite-Activate'}
    os = {'XR'}

    def run(self):

        """
        Produces a list of satellite ID that needs transfer

        RP/0/RP0/CPU0:AGN_PE_11_9k#install nv satellite 160,163 transfer
        """
        satellite_ids = self.ctx.load_job_data('selected_satellite_ids')
        pre_check_script = self.ctx.load_job_data('pre_check_script')[0]
        post_check_script = self.ctx.load_job_data('post_check_script')[0]

        if pre_check_script:
            pre_check_script = str(pre_check_script)
        if post_check_script:
            post_check_script = str(post_check_script)

        if pre_check_script:
            self.ctx.info("Satellite-Activate pre_check_script {} Pending".format(pre_check_script))
            try:
                output = subprocess.check_output(pre_check_script, shell=True)
                self.ctx.info("pre_check_script {} output:".format(pre_check_script))
                self.ctx.info("{}".format(output.decode('UTF-8').rstrip()))
            except subprocess.CalledProcessError as e:
                self.ctx.warning('Satellite-Activate pre_check_script {} error:'.format(pre_check_script))
                self.ctx.error(e.output)
            self.ctx.info("Satellite-Activate pre_check_script {} completed".format(pre_check_script))

        # ctype = type(satellite_ids[0])
        # self.ctx.info("ctype = {}".format(ctype))
        self.ctx.info("satellite_ids = {}".format(satellite_ids[0]))

        self.ctx.info("Satellite-Activate Pending")
        self.ctx.post_status("Satellite-Activate Pending")

        result = install_satellite_activate(self.ctx, satellite_ids[0])

        self.ctx.info("Refresh satellite inventory information")
        self.ctx.post_status("Refresh satellite inventory information")

        # Refresh satellite inventory information
        get_satellite(self.ctx)

        if post_check_script:
            self.ctx.info("Satellite-Activate post_check_script {} Pending".format(post_check_script))
            try:
                output = subprocess.check_output(post_check_script, shell=True)
                self.ctx.info("post_check_script {} output:".format(post_check_script))
                self.ctx.info("{}".format(output.decode('UTF-8').rstrip()))
            except subprocess.CalledProcessError as e:
                self.ctx.warning('Satellite-Activate post_check_script {} error:'.format(post_check_script))
                self.ctx.error(e.output)
            self.ctx.info("Satellite-Activate post_check_script {} completed".format(post_check_script))

        if result:
            self.ctx.info("Satellite-Activate completed")
        else:
            self.ctx.error("Satellite-Activate failed to complete.")



# coding=utf-8
# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re

plugin_ctx = None


def send_yes(fsm_ctx):
    fsm_ctx.ctrl.sendline('yes')
    return True


def send_newline(fsm_ctx):
    fsm_ctx.ctrl.sendline()
    return True


def fpd_package_installed(ctx):
    """
    :param ctx
    :return: True or False
    """

    active_packages = ctx.send("show install active summary")

    match = re.search("fpd", active_packages)

    if not match:
        return False
    else:
        return True


def fpd_needs_upgd(ctx, location, fpd_type):
    """
    :param ctx
    :param location
    :return: True or False

    Platform: ASR9000
    RP/0/RSP0/CPU0:cxr1#admin show hw-module fpd location 0/1/CPU0

    ===================================== ==========================================
                                          Existing Field Programmable Devices
                                          ==========================================
                                            HW                       Current SW Upg/
    Location     Card Type                Version Type Subtype Inst   Version   Dng?
    ============ ======================== ======= ==== ======= ==== =========== ====
    0/1/CPU0     A9K-MOD400-SE              1.0   lc   cbc     0      39.07     No
                                                  lc   rommon  0       8.43     No
                                                  lc   fpga2   0       1.91     Yes
                                                  lc   fsbl    0       1.96     Yes
                                                  lc   lnxfw   0       1.96     Yes
                                                  lc   fpga10  0       1.19     No
    --------------------------------------------------------------------------------
    NOTES:
    1.  One or more FPD needs an upgrade.  This can be accomplished
        using the "admin> upgrade hw-module fpd <fpd> location <loc>" CLI.
    """

    need_upgd = False
    sl = ['Subtype', 'Inst', 'Dng']
    dl = {}
    check = False
    cmd = 'admin show hw-module fpd location ' + location

    output = ctx.send(cmd)
    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:

        if '------------------------------------' in line or 'NOTES:' in line:
            break

        if line[0:8] == 'Location':
            for s in sl:
                n = line.find(s)
                if n != -1:
                    dl[str(s)] = n
                else:
                    ctx.error("unrecognized show hw-module fpd header {}".format(line))
                    return False
            continue

        if check:
            type = line[dl['Subtype']:dl['Inst']].strip()
            if type == fpd_type or fpd_type == 'all':
                if 'Yes' in line[dl['Dng']:]:
                    need_upgd = True
                    break
                else:
                    continue
            else:
                continue

        if line[0].isdigit():
            check = True
            type = line[dl['Subtype']:dl['Inst']].strip()
            if type == fpd_type or fpd_type == 'all':
                if 'Yes' in line[dl['Dng']:]:
                    need_upgd = True
                    break
                else:
                    continue
            else:
                continue

    return need_upgd


def fpd_locations(ctx):
    """
    :param ctx
    :return: A list of all locations

    Platform: ASR9000
    RP/0/RSP0/CPU0:cxr1#admin show hw-module fpd location all

    ===================================== ==========================================
                                          Existing Field Programmable Devices
                                          ==========================================
                                            HW                       Current SW Upg/
    Location     Card Type                Version Type Subtype Inst   Version   Dng?
    ============ ======================== ======= ==== ======= ==== =========== ====
    0/RSP0/CPU0  A9K-RSP440-TR              1.0   lc   cbc     0      16.116    No
                                                  lc   fpga2   0       1.10     No
                                                  lc   fpga3   0       4.09     No
                                                  lc   fpga1   0       0.11     No
                                                  lc   rommon  0       0.76     No
    --------------------------------------------------------------------------------
    0/FT0/SP     ASR-9904-FAN               1.0   ft   cbc     7      31.05     No
    --------------------------------------------------------------------------------
    0/BPID0/SP   ASR-9904-BPID2             1.0   bp   cbc     11      7.105    No
    --------------------------------------------------------------------------------
    0/PS0/M0/SP  PWR-3KW-AC-V2              1.0   pm   fpga14  13      3.18^    No
                                                  pm   fpga15  13      3.06^    No
                                                  pm   fpga16  13      3.12^    No
    --------------------------------------------------------------------------------
    0/PS0/M3/SP  PWR-3KW-AC-V2              1.0   pm   fpga14  16      3.18^    No
                                                  pm   fpga15  16      3.06^    No
                                                  pm   fpga16  16      3.12^    No
    --------------------------------------------------------------------------------
    0/RSP1/CPU0  A9K-RSP440-TR              1.0   lc   cbc     0      16.116    No
                                                  lc   fpga3   0       4.09     No
                                                  lc   fpga2   0       1.10     No
                                                  lc   fpga1   0       0.11     No
                                                  lc   rommon  0       0.76     No
    --------------------------------------------------------------------------------
    0/0/CPU0     A9K-MOD80-SE               1.0   lc   cbc     0      20.118    No
                                                  lc   fpga4   0       1.05     No
                                                  lc   fpga2   0       1.04     No
                                                  lc   rommon  0       3.03     No
    --------------------------------------------------------------------------------
    0/0/0        A9K-MPA-20X1GE             1.102 spa  fpga3   0       1.00     Yes
    --------------------------------------------------------------------------------
    0/1/CPU0     A9K-MOD400-SE              1.0   lc   cbc     0      39.07     No
                                                  lc   rommon  0       8.43     No
                                                  lc   fpga2   0       1.91     Yes
                                                  lc   fsbl    0       1.96     Yes
                                                  lc   lnxfw   0       1.96     Yes
                                                  lc   fpga10  0       1.19     No
    --------------------------------------------------------------------------------
    0/1/0        A9K-MPA-20X10GE            1.0   spa  fpga5   0       1.14     Yes
    --------------------------------------------------------------------------------
    NOTES:
    1.  One or more FPD needs an upgrade.  This can be accomplished
        using the "admin> upgrade hw-module fpd <fpd> location <loc>" CLI.
    2.  ^ One or more FPD will be intentionally skipped from upgrade using CLI with option "all" or during "Auto fpd".
          It can be upgraded only using the "admin> upgrade hw-module fpd <fpd> location <loc>" CLI with exact location.
    """

    sl = ['Location', 'Card Type']
    dl = {}
    locations = []

    output = ctx.send("admin show hw-module fpd location all")
    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:
        if line[0:8] == 'Location':
            for s in sl:
                n = line.find(s)
                if n != -1:
                    dl[str(s)] = n
                else:
                    ctx.error("unrecognized admin show hw-module fpd location header {}".format(line))
                    return False

        if line[0].isdigit():
            location = line[dl['Location']:dl['Card Type']].strip()
            locations.append(location)

        if 'NOTES:' in line:
            break

    return locations


def hw_fpd_upgd(ctx, location, type):
    """
    :param ctx
    :param location
    :return: True or False

    Platform: ASR9000
    RP/0/RSP0/CPU0:vkg1#admin upgrade hw-module fpd all location 0/RSP0/CPU0
    Wed Jul  1 01:57:15.137 UTC

    ***** UPGRADE WARNING MESSAGE: *****
      *  This upgrade operation has a maximum timout of 160 minutes.  *
      *  If you are executing the cmd for one specific location and  *
      *  card in that location reloads or goes down for some reason  *
      *  you can press CTRL-C to get back the RP's prompt.           *
      *  If you are executing the cmd for _all_ locations and a node *
      *  reloads or is down please allow other nodes to finish the   *
      *  upgrade process before pressing CTRL-C.                     *

    % RELOAD REMINDER:
      - The upgrade operation of the target module will not interrupt its normal
        operation. However, for the changes to take effect, the target module
        will need to be manually reloaded after the upgrade operation. This can
        be accomplished with the use of "hw-module <target> reload" command.
      - If automatic reload operation is desired after the upgrade, please use
        the "reload" option at the end of the upgrade command.
      - The output of "show hw-module fpd location" command will not display
        correct version information after the upgrade if the target module is
        not reloaded.
    NOTE: Chassis CLI will not be accessible while upgrade is in progress.
    Continue? [confirm]


    FPD upgrade in progress on some hardware, reload/configuration change
    on those is not recommended as it might cause HW programming failure
    and result in RMA of the hardware.


    Starting the upgrade/download of following FPDs:
    =========== ==== ======= ======= =========== =========
                                       Current    Upg/Dng
    Location    Type Subtype Upg/Dng   Version    Version
    =========== ==== ======= ======= =========== =========
    0/RSP0/CPU0 lc   fpga2   upg         1.06        1.09
                lc   fpga    upg         0.10        0.11
                lc   rommon  upg         0.71        0.75
    ------------------------------------------------------
    FPD upgrade in progress. Max timeout remaining 89 min.
    FPD upgrade in progress. Max timeout remaining 88 min.
    FPD upgrade in progress. Max timeout remaining 87 min.
    FPD upgrade in progress. Max timeout remaining 86 min.
    FPD upgrade in progress. Max timeout remaining 85 min.
    FPD upgrade in progress. Max timeout remaining 84 min.
    FPD upgrade in progress. Max timeout remaining 83 min.
    FPD upgrade in progress. Max timeout remaining 82 min.
    FPD upgrade in progress. Max timeout remaining 81 min.
    FPD upgrade in progress. Max timeout remaining 80 min.
    Successfully upgraded     fpga2 for        A9K-RSP440-TR on location 0/RSP0/CPU0 from  1.06 to  1.09
    Successfully upgraded      fpga for        A9K-RSP440-TR on location 0/RSP0/CPU0 from  0.10 to  0.11
    Successfully upgraded    rommon for        A9K-RSP440-TR on location 0/RSP0/CPU0 from  0.71 to  0.75


    FPD upgrade has ended.

    RP/0/RSP0/CPU0:RO#admin upgrade hw-module fpd all location 0/RSP0/CPU0
    Fri Jun 30 16:16:27.326 PST

    ***** UPGRADE WARNING MESSAGE: *****
      *  This upgrade operation has a maximum timout of 90 minutes.  *
      *  If you are executing the cmd for one specific location and  *
      *  card in that location reloads or goes down for some reason  *
      *  you can press CTRL-C to get back the RP's prompt.           *
      *  If you are executing the cmd for _all_ locations and a node *
      *  reloads or is down please allow other nodes to finish the   *
      *  upgrade process before pressing CTRL-C.                     *

    % RELOAD REMINDER:
      - The upgrade operation of the target module will not interrupt its normal
        operation. However, for the changes to take effect, the target module
        will need to be manually reloaded after the upgrade operation. This can
        be accomplished with the use of "hw-module <target> reload" command.
      - If automatic reload operation is desired after the upgrade, please use
        the "reload" option at the end of the upgrade command.
      - The output of "show hw-module fpd location" command will not display
        correct version information after the upgrade if the target module is
        not reloaded.
    NOTE: Chassis CLI will not be accessible while upgrade is in progress.
    Continue ? [no]: yes



    FPD upgrade in progress on some hardware, reload/configuration change
    on those is not recommended as it might cause HW programming failure
    and result in RMA of the hardware.


    Starting the upgrade/download of following FPDs:
    =========== ==== ======= ======= =========== =========
                                       Current    Upg/Dng
    Location    Type Subtype Upg/Dng   Version    Version
    =========== ==== ======= ======= =========== =========
    0/RSP0/CPU0 lc   fpga2   upg         1.09        1.10
                lc   rommon  upg         0.75        0.76
    ------------------------------------------------------
    FPD upgrade in progress. Max timeout remaining 89 min.
    FPD upgrade in progress. Max timeout remaining 88 min.
    Successfully upgraded     fpga2 for        A9K-RSP440-SE on location 0/RSP0/CPU0 from  1.09 to  1.10
    Successfully upgraded    rommon for        A9K-RSP440-SE on location 0/RSP0/CPU0 from  0.75 to  0.76


    FPD upgrade has ended.
    """
    global plugin_ctx
    plugin_ctx = ctx

    cmd = 'admin upgrade hw-module fpd ' + type + ' location ' + location

    # There can be two different prompts
    CONTINUE_YES = re.compile(r"Continue\? \[confirm\]")
    CONTINUE_UPGD = re.compile(r"Continue \? \[no\]:")
    END_UPGD = re.compile(r"FPD upgrade has ended.")
    COMPLETE_UPGD = re.compile(r"FPD upgrade completed. Auto-reload triggered")
    HOST_PROMPT = re.compile(ctx.prompt)

    events = [CONTINUE_YES, CONTINUE_UPGD, END_UPGD, COMPLETE_UPGD, HOST_PROMPT]
    transitions = [
        (CONTINUE_YES, [0], 1, send_newline, 9600),
        (CONTINUE_UPGD, [0, 1], 2, send_yes, 9600),
        (END_UPGD, [1, 2], 3, None, 9600),
        (COMPLETE_UPGD, [1, 2], 3, None, 9600),
        (HOST_PROMPT, [3], -1, None, 9600)
    ]

    if not ctx.run_fsm("Upgrade hw-module fpd", cmd, events, transitions, timeout=9600):
        ctx.warning("Failed: {}".format(cmd))
        return False

    return True


def fpd_check_status(ctx, location, type):
    """
    :param ctx
    :param locations
    :return: True or False

    Platform: asr9000
    RP/0/RSP0/CPU0:cxr1#admin show hw-module fpd location all
    Fri Jun 30 17:57:53.343 UTC

    ===================================== ==========================================
                                          Existing Field Programmable Devices
                                          ==========================================
                                            HW                       Current SW Upg/
    Location     Card Type                Version Type Subtype Inst   Version   Dng?
    ============ ======================== ======= ==== ======= ==== =========== ====
    0/RSP0/CPU0  A9K-RSP440-TR              1.0   lc   cbc     0      16.116    No
                                                  lc   fpga1   0       0.11     No
                                                  lc   fpga3   0       4.09     No
                                                  lc   fpga2   0       1.10     No
                                                  lc   rommon  0       0.76     No
    --------------------------------------------------------------------------------
    0/FT0/SP     ASR-9904-FAN               1.0   ft   cbc     7      31.05     No
    --------------------------------------------------------------------------------
    0/BPID0/SP   ASR-9904-BPID2             1.0   bp   cbc     11      7.105    No
    --------------------------------------------------------------------------------
    0/PS0/M0/SP  PWR-3KW-AC-V2              1.0   pm   fpga14  13      3.18^    No
                                                  pm   fpga15  13      3.06^    No
                                                  pm   fpga16  13      3.12^    No
    --------------------------------------------------------------------------------
    0/PS0/M3/SP  PWR-3KW-AC-V2              1.0   pm   fpga14  16      3.18^    No
                                                  pm   fpga15  16      3.06^    No
                                                  pm   fpga16  16      3.12^    No
    --------------------------------------------------------------------------------
    0/RSP1/CPU0  A9K-RSP440-TR              1.0   lc   cbc     0      16.116    No
                                                  lc   fpga1   0       0.11     No
                                                  lc   fpga2   0       1.10     No
                                                  lc   fpga3   0       4.09     No
                                                  lc   rommon  0       0.76     No
    --------------------------------------------------------------------------------
    0/0/CPU0     A9K-MOD80-SE               1.0   lc   cbc     0      20.118    No
                                                  lc   fpga2   0       1.04     No
                                                  lc   fpga4   0       1.05     No
                                                  lc   rommon  0       3.03     No
    --------------------------------------------------------------------------------
    0/0/0        A9K-MPA-20X1GE             1.102 spa  fpga3   0       1.01     No
    --------------------------------------------------------------------------------
    0/1/CPU0     A9K-MOD400-SE              1.0   lc   cbc     0      39.07     No
                                                  lc   rommon  0       8.43     No
                                                  lc   fpga2   0       1.93     No
                                                  lc   fsbl    0       1.100    No
                                                  lc   lnxfw   0       1.100    No
                                                  lc   fpga10  0       1.19     No
    --------------------------------------------------------------------------------
    0/1/0        A9K-MPA-20X10GE            1.0   spa  fpga5   0       1.16     No
    --------------------------------------------------------------------------------
    NOTES:
    1.  ^ One or more FPD will be intentionally skipped from upgrade using CLI with option "all" or during "Auto fpd".
          It can be upgraded only using the "admin> upgrade hw-module fpd <fpd> location <loc>" CLI with exact location.
    """

    upgd_result = True
    if fpd_needs_upgd(ctx, location, type):
        upgd_result = False
        cmd = 'admin show hw-module fpd location ' + location
        output = ctx.send(cmd)
        ctx.warning("FPD Upgrade result for {}".format(cmd))
        ctx.warning("{}".format(output))

    return upgd_result


def hw_fpd_reload(ctx, location):
    """
    :param ctx
    :param location:
    :return: True or False

    Platform: ASR9000
    RP/0/RSP0/CPU0:RP2#admin reload location all
    Mon Jun 29 20:59:37.902 PST
    Some active software packages are not yet committed. Proceed?[confirm]

    Preparing system for backup. This may take a few minutes especially for large configurations.
            Status report: node0_RSP0_CPU0: START TO BACKUP
            Status report: node0_RSP0_CPU0: BACKUP HAS COMPLETED SUCCESSFULLY
    [Done]
    Proceed with reload? [confirm]RP/0/RSP0/CPU0::This node received reload command. Reloading in 5 secs
    """
    global plugin_ctx
    plugin_ctx = ctx

    cmd = 'admin reload location ' + location
    fsm_name = 'Reload location ' + location

    # Seeing this message without the reboot prompt indicates a non-reload situation
    PROCEED_RELOAD = re.compile(r"software packages are not yet committed. Proceed\?\[confirm\]")
    DONE = re.compile(re.escape(r"[Done]"))
    PROCEED = re.compile(re.escape(r"Proceed with reload? [confirm]"))

    events = [PROCEED_RELOAD, DONE, PROCEED]
    transitions = [
        (PROCEED_RELOAD, [0], 1, send_newline, 300),
        (DONE, [0, 1], 2, None, 300),
        (PROCEED, [0, 1, 2], -1, send_newline, 300)
    ]

    if not ctx.run_fsm(fsm_name, cmd, events, transitions, timeout=300):
        ctx.warning("Failed: {}".format(cmd))
        return False

    return True


def active_rsp_location(ctx):
    """
    :param ctx
    :return: the location of the active RSP / RP
    """

    location = ''
    cmd = 'show platform | include Active'
    output = ctx.send(cmd)

    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:
        if line[0].isdigit():
            location = line[0:16].strip()

    return location


def cbc_pwr_only(ctx):
    """
    :param ctx
    :return: True or False

    Platform: ASR9000
    RP/0/RSP0/CPU0:cxr1#admin show hw-module fpd location all

    ===================================== ==========================================
                                          Existing Field Programmable Devices
                                          ==========================================
                                            HW                       Current SW Upg/
    Location     Card Type                Version Type Subtype Inst   Version   Dng?
    ============ ======================== ======= ==== ======= ==== =========== ====
    0/RSP0/CPU0  A9K-RSP440-TR              1.0   lc   cbc     0      16.116    No
                                                  lc   fpga2   0       1.10     No
                                                  lc   fpga3   0       4.09     No
                                                  lc   fpga1   0       0.11     No
                                                  lc   rommon  0       0.76     No
    --------------------------------------------------------------------------------
    0/FT0/SP     ASR-9904-FAN               1.0   ft   cbc     7      31.05     No
    --------------------------------------------------------------------------------
    0/BPID0/SP   ASR-9904-BPID2             1.0   bp   cbc     11      7.105    No
    --------------------------------------------------------------------------------
    0/PS0/M0/SP  PWR-3KW-AC-V2              1.0   pm   fpga14  13      3.18^    No
                                                  pm   fpga15  13      3.06^    No
                                                  pm   fpga16  13      3.12^    No
    --------------------------------------------------------------------------------
    0/PS0/M3/SP  PWR-3KW-AC-V2              1.0   pm   fpga14  16      3.18^    No
                                                  pm   fpga15  16      3.06^    No
                                                  pm   fpga16  16      3.12^    No
    --------------------------------------------------------------------------------
    0/RSP1/CPU0  A9K-RSP440-TR              1.0   lc   cbc     0      16.116    No
                                                  lc   fpga3   0       4.09     No
                                                  lc   fpga2   0       1.10     No
                                                  lc   fpga1   0       0.11     No
                                                  lc   rommon  0       0.76     No
    --------------------------------------------------------------------------------
    0/0/CPU0     A9K-MOD80-SE               1.0   lc   cbc     0      20.118    No
                                                  lc   fpga4   0       1.05     No
                                                  lc   fpga2   0       1.04     No
                                                  lc   rommon  0       3.03     No
    --------------------------------------------------------------------------------
    0/0/0        A9K-MPA-20X1GE             1.102 spa  fpga3   0       1.00     Yes
    --------------------------------------------------------------------------------
    0/1/CPU0     A9K-MOD400-SE              1.0   lc   cbc     0      39.07     No
                                                  lc   rommon  0       8.43     No
                                                  lc   fpga2   0       1.91     Yes
                                                  lc   fsbl    0       1.96     Yes
                                                  lc   lnxfw   0       1.96     Yes
                                                  lc   fpga10  0       1.19     No
    --------------------------------------------------------------------------------
    0/1/0        A9K-MPA-20X10GE            1.0   spa  fpga5   0       1.14     Yes
    --------------------------------------------------------------------------------
    NOTES:
    1.  One or more FPD needs an upgrade.  This can be accomplished
        using the "admin> upgrade hw-module fpd <fpd> location <loc>" CLI.
    2.  ^ One or more FPD will be intentionally skipped from upgrade using CLI with option "all" or during "Auto fpd".
          It can be upgraded only using the "admin> upgrade hw-module fpd <fpd> location <loc>" CLI with exact location.
    """

    avoid_reload = True
    sl = ['Type', 'Subtype', 'Inst', 'Dng']
    dl = {}
    check = False
    cmd = 'admin show hw-module fpd location all'

    output = ctx.send(cmd)
    lines = output.split('\n')
    lines = [x for x in lines if x]
    for line in lines:

        if '------------------------------------' in line:
            continue

        if 'NOTES:' in line:
            break

        if line[0:8] == 'Location':
            for s in sl:
                n = line.find(s)
                if n != -1:
                    dl[str(s)] = n
                else:
                    ctx.warning("unrecognized show hw-module fpd header {}".format(line))
                    return False
            continue

        if check:
            subtype = line[dl['Subtype']:dl['Inst']].strip()
            type = line[dl['Type']:dl['Subtype']].strip()
            if 'pm' in type:
                continue
            if 'Yes' in line[dl['Dng']:] and 'cbc' not in subtype:
                avoid_reload = False
                break
            else:
                continue

        if line[0].isdigit():
            check = True
            subtype = line[dl['Subtype']:dl['Inst']].strip()
            type = line[dl['Type']:dl['Subtype']].strip()
            if 'pm' in type:
                continue
            if 'Yes' in line[dl['Dng']:] and 'cbc' not in subtype:
                avoid_reload = False
                break
            else:
                continue

    return avoid_reload



# =============================================================================
#
# Copyright (c) 2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from csmpe.plugins import CSMPlugin
from install import install_activate_deactivate, parse_pkg_list, report_changed_pkg
from csmpe.core_plugins.csm_get_inventory.ios_xr.plugin import get_package, get_inventory
from csmpe.core_plugins.csm_install_operations.utils import update_device_info_udi


class Plugin(CSMPlugin):
    """This plugin rollback to the last committed installation point."""
    name = "Install Rollback Plugin"
    platforms = {'ASR9K', 'XR12K', 'CRS'}
    phases = {'Rollback'}
    os = {'XR'}

    def run(self):

        self.ctx.send("admin show install active summary")
        self.ctx.send("admin show install committed summary")

        cmd = 'admin install rollback to committed prompt-level none asynchronous'

        output = self.ctx.send("show install active summary")
        before = parse_pkg_list(output)

        self.ctx.info("Install rollback to the last committed installation point Pending")
        self.ctx.post_status("Install rollback to the last committed installation point Pending")

        install_activate_deactivate(self.ctx, cmd)

        self.ctx.info("Install rollback done")

        output = self.ctx.send("show install active summary")
        after = parse_pkg_list(output)

        pkg_list = report_changed_pkg(before, after)
        pkg_change_list = ','.join(pkg_list)

        self.ctx.save_job_data("package_change_list", pkg_change_list)
        self.ctx.info("package change list:")

        for pkg in pkg_list:
            self.ctx.info("\t{}".format(pkg))

        self.ctx.info("Refreshing package and inventory information")
        self.ctx.post_status("Refreshing package and inventory information")

        # Refresh package and inventory information
        get_package(self.ctx)
        get_inventory(self.ctx)

        update_device_info_udi(self.ctx)
